# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

"""
Embedded-atom method potential.
"""

import os

import numpy as np

import ase
from ase.calculators.calculator import Calculator

try:
    from scipy.interpolate import InterpolatedUnivariateSpline
except:
    InterpolatedUnivariateSpline = None

from matscipy.calculators.eam.io import read_eam
from matscipy.neighbours import neighbour_list

###

def _make_splines(dx, y):
    if len(np.asarray(y).shape) > 1:
        return [_make_splines(dx, yy) for yy in y]
    else:
        return InterpolatedUnivariateSpline(np.arange(len(y))*dx, y)

def _make_derivative(x):
    if type(x) == list:
        return [_make_derivative(xx) for xx in x]
    else:
        return x.derivative()

###

class EAM(Calculator):
    implemented_properties = ['energy', 'stress', 'forces']
    default_parameters = {}
    name = 'EAM'
       
    def __init__(self, fn=None, atomic_numbers=None, F=None, f=None, rep=None,
                 cutoff=None, kind='eam/alloy'):
        Calculator.__init__(self)
        if fn is not None:
            source, parameters, F, f, rep = read_eam(fn, kind=kind)
            self.atnums = parameters.atomic_numbers
            self.cutoff = parameters.cutoff
            dr = parameters.distance_grid_spacing
            dF = parameters.density_grid_spacing

            # Create spline interpolation
            self.F = _make_splines(dF, F)
            self.f = _make_splines(dr, f)
            self.rep = _make_splines(dr, rep)
        else:
            self.atnums = atomic_numbers
            self.F = F
            self.f = f
            self.rep = rep
            self.cutoff = cutoff

        self.atnum_to_index = -np.ones(np.max(self.atnums)+1, dtype=int)
        self.atnum_to_index[self.atnums] = \
            np.arange(len(self.atnums))

        # Derivative of spline interpolation
        self.dF = _make_derivative(self.F)
        self.df = _make_derivative(self.f)
        self.drep = _make_derivative(self.rep)

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        nat = len(self.atoms)
        atnums = self.atoms.numbers

        atnums_in_system = set(atnums)
        for atnum in atnums_in_system:
            if atnum not in self.atnums:
                raise RuntimeError('Element with atomic number {} found, but '
                                   'this atomic number has no EAM '
                                   'parametrization'.format(atnum))

        i_n, j_n, dr_nc, abs_dr_n = neighbour_list('ijDd', self.atoms,
                                                   self.cutoff)

        # Density
        f_n = np.zeros_like(abs_dr_n)
        df_n = np.zeros_like(abs_dr_n)
        for atidx1, atnum1 in enumerate(self.atnums):
            f1 = self.f[atidx1]
            df1 = self.df[atidx1]
            mask1 = atnums[j_n]==atnum1
            if mask1.sum() > 0:
                if type(f1) == list:
                    for atidx2, atnum2 in enumerate(self.atnums):
                        f = f1[atidx2]
                        df = df1[atidx2]
                        mask = np.logical_and(mask1, atnums[i_n]==atnum2)
                        if mask.sum() > 0:
                            f_n[mask] = f(abs_dr_n[mask])
                            df_n[mask] = df(abs_dr_n[mask])
                else:
                    f_n[mask1] = f1(abs_dr_n[mask1])
                    df_n[mask1] = df1(abs_dr_n[mask1])

        density_i = np.bincount(i_n, weights=f_n, minlength=nat)

        # Repulsion
        rep_n = np.zeros_like(abs_dr_n)
        drep_n = np.zeros_like(abs_dr_n)
        for atidx1, atnum1 in enumerate(self.atnums):
            rep1 = self.rep[atidx1]
            drep1 = self.drep[atidx1]
            mask1 = atnums[i_n]==atnum1
            if mask1.sum() > 0:
                for atidx2, atnum2 in enumerate(self.atnums):
                    rep = rep1[atidx2]
                    drep = drep1[atidx2]
                    mask = np.logical_and(mask1, atnums[j_n]==atnum2)
                    if mask.sum() > 0:
                        r = rep(abs_dr_n[mask])/abs_dr_n[mask]
                        rep_n[mask] = r
                        drep_n[mask] = (drep(abs_dr_n[mask])-r)/abs_dr_n[mask]

        # Energy
        epot = 0.5*np.sum(rep_n)
        demb_i = np.zeros(len(self.atoms))
        for atidx, atnum in enumerate(self.atnums):
            F = self.F[atidx]
            dF = self.dF[atidx]
            mask = atnums==atnum
            if mask.sum() > 0:
                epot += np.sum(F(density_i[mask]))
                demb_i[mask] += dF(density_i[mask])

        # Forces
        df_nc = -0.5*((demb_i[i_n]+demb_i[j_n])*df_n+drep_n).reshape(-1,1)*dr_nc/abs_dr_n.reshape(-1,1)

        # Sum for each atom
        fx_i = np.bincount(j_n, weights=df_nc[:,0], minlength=nat) - \
            np.bincount(i_n, weights=df_nc[:,0], minlength=nat)
        fy_i = np.bincount(j_n, weights=df_nc[:,1], minlength=nat) - \
            np.bincount(i_n, weights=df_nc[:,1], minlength=nat)
        fz_i = np.bincount(j_n, weights=df_nc[:,2], minlength=nat) - \
            np.bincount(i_n, weights=df_nc[:,2], minlength=nat)

        # Virial
        virial_v = -np.array([dr_nc[:,0]*df_nc[:,0],               # xx
                              dr_nc[:,1]*df_nc[:,1],               # yy
                              dr_nc[:,2]*df_nc[:,2],               # zz
                              dr_nc[:,1]*df_nc[:,2],               # yz
                              dr_nc[:,0]*df_nc[:,2],               # xz
                              dr_nc[:,0]*df_nc[:,1]]).sum(axis=1)  # xy

        self.results = {'energy': epot,
                        'stress': virial_v/self.atoms.get_volume(),
                        'forces': np.transpose([fx_i, fy_i, fz_i])}



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import math

import numpy as np

###

def radius(N, R, Es, w):
    """
    Given normal load, sphere radius and contact modulus compute contact radius
    and peak pressure.

    Parameters
    ----------
    N : float
        Normal force.
    R : float
        Sphere radius.
    Es : float
        Contact modulus: Es = E/(1-nu**2) with Young's modulus E and Poisson
        number nu.
    w : float
        Work of adhesion.
    """

    return ( 3.*R/(4.*Es)*(N + 6*w*math.pi*R + np.sqrt(12*w*math.pi*R*N + 
                           (6*w*math.pi*R)**2)) )**(1./3)




# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

"""
Log status to screen.
"""

from __future__ import print_function

import os
import sys

import inspect

from functools import reduce
from math import isnan, isinf
from numbers import Real

###

def hdr_str(s, x):
    """ Return header description strings
    """

    if isinstance(x, str):
        return s

    r = [ ]
    try:
        for i, v in enumerate(x):
            r += [ s+'('+chr(ord('x')+i)+')' ]
    except:
        r = s
    return r

def hdrfmt_str(x, i):
    """ Return header format string for datatype x
    """

    if isinstance(x, str):
        return '{'+str(i)+':>20}'
    elif isinstance(x, int):
        return '{'+str(i)+':>8}'
    elif isinstance(x, Real):
        return '{'+str(i)+':>20}'
    else:
        # Is this something we need to iterate over?
        try:
            return [ hdrfmt_str(k,i+j) for j, k in enumerate(x) ]
        except:
            return '{'+str(i)+':>20}'
    
def numfmt_str(x, i):
    """ Return numeric format string for datatype x
    """

    if isinstance(x, str):
        return '{'+str(i)+':>20}'
    elif isinstance(x, int):
        return '{'+str(i)+':>8}'
    elif isinstance(x, Real):
        return '{'+str(i)+':>20.12e}'
    else:
        # Is this something we need to iterate over?
        try:
            return [ numfmt_str(k,i+j) for j, k in enumerate(x) ]
        except:
            return '{'+str(i)+':>20}'

def flatten(x):
    if isinstance(x, str):
        return [ x ]
    else:
        # Is this something we can iterate over?
        try:
            return reduce(lambda a,b: a+b, [ flatten(i) for i in x ])
        except:
            return [ x ]

###

class Logger(object):
    # Debug option, redirect all output to screen
    __all_output_to_stdout = False

    def __init__(self, logfile=sys.stdout, outevery=1, sepevery=10):
        self.sepevery = sepevery

        self.set_outevery(outevery)

        self.it = 1
        self.logfn = None
        self.logfile = None

        self.buffer = [ ]

        self.set_logfile(logfile)


    def __open_logfile(self):
        if self.logfile is None and self.logfn is not None and \
                not self.__all_output_to_stdout:
            self.outcounter = self.outevery
            self.sepcounter = 0
            fn = self.logfn.format(self.it)
            if os.path.exists(fn):
                # Save old log file as .bak
                i = 0
                while os.path.exists('{0}.{1}.bak'.format(fn, i)):
                    i += 1
                os.rename(fn, '{0}.{1}.bak'.format(fn, i))
            self.logfile = open(fn, 'w')


    def _print(self, s, logfile=None):
        if logfile and self.logfile != logfile:
            print(s, file=logfile)
        if self.logfile:
            print(s, file=self.logfile)
        else:
            self.buffer += [ s ]


    def flush(self):
        if self.logfile:
            self.logfile.flush()


    def set_logfile(self, logfile):
        if self.__all_output_to_stdout:
            self.logfile = sys.stdout
        elif isinstance(logfile, str):
            if logfile.find('{0}') != -1 or logfile.find('{}') != -1:
                self.logfn = logfile
                self.__open_logfile()
            else:
                self.logfile = open(logfile, 'w')
        else:
            self.logfile = logfile

        if self.logfile is not None:
            for s in self.buffer:
                self._print(s)

        self.buffer = [ ]


    def pr(self, s, caller=None, logfile=None):
        self.__open_logfile()
        if caller is None:
            caller = inspect.stack()[1]
        self._print('# {{{0}}}: {1}'.format(caller[3], s), logfile=logfile)
        self.flush()


    def warn(self, s, caller=None):
        self.pr('Warning: '+s, caller=caller, logfile=sys.stdout)


    def st(self, hdr, vals, force_print=False):
        assert len(hdr) == len(vals)
        self.__open_logfile()

        do_print = force_print

        if self.outevery == 1:
            do_print = True
        else:
            self.outcounter -= 1
            if self.outcounter <= 0:
                do_print = True
                self.outcounter = self.outevery
        
        if do_print:
            self.sepcounter -= 1
            if self.sepcounter <= 0:
                # For vectors we need a column for each component
                hdr = flatten([ hdr_str(a, b) for a,b in zip(hdr, vals) ])
                fmt_str = '#'+reduce(
                    lambda a,b: '{0}  {1}'.format(a, b),
                    flatten([ hdrfmt_str(i,j)
                              for j,i in enumerate(flatten(vals)) ])
                    )
                self._print(fmt_str.format(*[ '{0}:{1}'.format(str(i+1), s)
                                              for i, s in enumerate(hdr) ]))
                self.sepcounter = self.sepevery

            fmt_str = ' '+reduce(
                lambda a,b: '{0}  {1}'.format(a, b),
                flatten([ numfmt_str(i,j)
                          for j,i in enumerate(flatten(vals)) ])
                )
            self._print(fmt_str.format(*flatten(vals)))
            self.flush()


    def iteration_finished(self):
        self.it += 1
        self.outcounter = self.outevery
        self.sepcounter = 0
        if self.logfn is not None:
            self.logfile = None


    def get_logfile(self):
        return self.logfile


    def has_logfile(self):
        return self.logfile is not None


    def set_outevery(self, outevery):
        self.outevery = outevery

        self.outcounter = outevery
        self.sepcounter = 0

###

quiet = Logger(None)
screen = Logger()



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

from __future__ import print_function

import numpy as np

from ase.lattice.cubic import Diamond, FaceCenteredCubic, SimpleCubic

###

def set_groups(a, n, skin_x, skin_y, central_x=-1./2, central_y=-1./2,
               invert_central=False):
    nx, ny, nz = n
    sx, sy, sz = a.cell.diagonal()
    print('skin_x = {0}*a0, skin_y = {1}*a0'.format(skin_x, skin_y))
    skin_x = skin_x*sx/nx
    skin_y = skin_y*sy/ny
    print('skin_x = {0}, skin_y = {1}'.format(skin_x, skin_y))
    r = a.positions

    g = np.ones(len(a), dtype=int)
    mask = np.logical_or(
               np.logical_or(
                   np.logical_or(
                       r[:, 0]/sx < (1.-central_x)/2,
                       r[:, 0]/sx > (1.+central_x)/2),
                   r[:, 1]/sy < (1.-central_y)/2),
               r[:, 1]/sy > (1.+central_y)/2)
    if invert_central:
        mask = np.logical_not(mask)
    g = np.where(mask, g, 2*np.ones_like(g))

    mask = np.logical_or(
               np.logical_or(
                   np.logical_or(
                       r[:, 0] < skin_x, r[:, 0] > sx-skin_x),
                   r[:, 1] < skin_y),
               r[:, 1] > sy-skin_y)
    g = np.where(mask, np.zeros_like(g), g)

    a.set_array('groups', g)


def cluster(el, a0, n, crack_surface=[1,1,0], crack_front=[0,0,1],
            lattice=None, shift=None):
    nx, ny, nz = n
    third_dir = np.cross(crack_surface, crack_front)
    directions = [ third_dir, crack_surface, crack_front ]
    if np.linalg.det(directions) < 0:
        third_dir = -third_dir
    directions = [ third_dir, crack_surface, crack_front ]
    unitcell = lattice(el, latticeconstant=a0, size=[1, 1, 1], 
                       directions=directions  )
    if shift is not None:
        unitcell.translate(np.dot(shift, unitcell.cell))

    # Center cluster in unit cell
    x, y, z = (unitcell.get_scaled_positions()%1.0).T
    x += (1.0-x.max()+x.min())/2 - x.min()
    y += (1.0-y.max()+y.min())/2 - y.min()
    z += (1.0-z.max()+z.min())/2 - z.min()
    unitcell.set_scaled_positions(np.transpose([x, y, z]))

    a = unitcell.copy()
    a *= (nx, ny, nz)
    #a.info['unitcell'] = unitcell

    a.set_pbc([False, False, True])

    return a


def diamond(*args, **kwargs):
    kwargs['lattice'] = Diamond
    return cluster(*args, **kwargs)


def fcc(*args, **kwargs):
    kwargs['lattice'] = FaceCenteredCubic
    return cluster(*args, **kwargs)


def sc(*args, **kwargs):
    kwargs['lattice'] = SimpleCubic
    return cluster(*args, **kwargs)



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import math

import numpy as np

import matscipy.contact_mechanics.Hertz as Hertz

###

def radius_and_pressure(N, R, Es, w):
    """
    Given normal load, sphere radius and contact modulus compute contact radius
    and peak pressure.

    Parameters
    ----------
    N : float
        Normal force.
    R : float
        Sphere radius.
    Es : float
        Contact modulus: Es = E/(1-nu**2) with Young's modulus E and Poisson
        number nu.
    w : float
        Work of adhesion.
    """

    return Hertz.radius_and_pressure(N+2*math.pi*w*R, R, Es)


def surface_stress(r, a, nu):
    """
    Given distance from the center of the sphere, contact radius and Poisson
    number contact, compute the stress at the surface.

    Parameters
    ----------
    r : array_like
        Array of distance (from the center of the sphere).
    a : float
        Contact radius.
    nu : float
        Poisson number.

    Returns
    -------
    pz : array
        Contact pressure.
    sr : array
        Radial stress.
    stheta : array
        Azimuthal stress.
    """

    return Hertz.surface_stress(r, a, nu)
    
    
def surface_displacements(r, a):
    return Hertz.surface_displacements(r, a)


# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014-2018) James Kermode, King's College London
#                       Lars Pastewka, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

"""
Simple pair potential.
"""

from __future__ import division

import os

import sys

import time

import numpy as np

import ase
from ase.calculators.calculator import Calculator

from matscipy.neighbours import neighbour_list, first_neighbours


###

class LennardJonesCut():
    """
    Functional form for a 12-6 Lennard-Jones potential with a hard cutoff.
    Energy is shifted to zero at cutoff.
    """

    def __init__(self, epsilon, sigma, cutoff):
        self.epsilon = epsilon
        self.sigma = sigma
        self.cutoff = cutoff
        self.offset = (sigma/cutoff)**12 - (sigma/cutoff)**6

    def __call__(self, r):
        """
        Return function value (potential energy).
        """
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((r6-1) * r6 - self.offset)

    def get_cutoff(self):
        return self.cutoff

    def first_derivative(self, r):
        r = (self.sigma / r)
        r6 = r**6
        return -24 * self.epsilon / self.sigma * (2*r6-1) * r6 * r

    def second_derivative(self, r):
        r2 = (self.sigma / r)**2
        r6 = r2**3
        return 24 * self.epsilon/self.sigma**2 * (26*r6-7) * r6 * r2

    def derivative(self, n=1):
        if n == 1:
            return self.first_derivative
        elif n == 2:
            return self.second_derivative
        else:
            raise ValueError(
                "Don't know how to compute {}-th derivative.".format(n))

###


class LennardJonesQuadratic():
    """
    Functional form for a 12-6 Lennard-Jones potential with a soft cutoff.
    Energy, its first and second derivative are shifted to zero at cutoff.
    """

    def __init__(self, epsilon, sigma, cutoff):
        self.epsilon = epsilon
        self.sigma = sigma
        self.cutoff = cutoff
        self.offset_energy = (sigma/cutoff)**12 - (sigma/cutoff)**6
        self.offset_force = 6/cutoff * \
            (-2*(sigma/cutoff)**12+(sigma/cutoff)**6)
        self.offset_dforce = (1/cutoff**2) * \
            (156*(sigma/cutoff)**12-42*(sigma/cutoff)**6)

    def get_cutoff(self):
        return self.cutoff

    def __call__(self, r):
        """
        Return function value (potential energy).
        """
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((r6-1)*r6-self.offset_energy - (r-self.cutoff) * self.offset_force - ((r - self.cutoff)**2/2) * self.offset_dforce)

    def first_derivative(self, r):
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((6/r) * (-2*r6+1) * r6 - self.offset_force - (r-self.cutoff) * self.offset_dforce)

    def second_derivative(self, r):
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((1/r**2) * (156*r6-42) * r6 - self.offset_dforce)

    def derivative(self, n=1):
        if n == 1:
            return self.first_derivative
        elif n == 2:
            return self.second_derivative
        else:
            raise ValueError(
                "Don't know how to compute {}-th derivative.".format(n))

###


class LennardJonesLinear():
    """
    Function form of a 12-6 Lennard-Jones potential with a soft cutoff
    The energy and the force are shifted at the cutoff.
    """

    def __init__(self, epsilon, sigma, cutoff):
        self.epsilon = epsilon
        self.sigma = sigma
        self.cutoff = cutoff
        self.offset_energy = (sigma/cutoff)**12 - (sigma/cutoff)**6
        self.offset_force = 6/cutoff * \
            (-2*(sigma/cutoff)**12+(sigma/cutoff)**6)

    def get_cutoff():
        return self.cutoff

    def __call__(self, r):
        """
        Return function value (potential energy).
        """
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((r6-1) * r6 - self.offset_energy - (r-self.cutoff) * self.offset_force)

    def first_derivative(self, r):
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((6/r) * (-2*r6+1) * r6 - self.offset_force)

    def second_derivative(self, r):
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((1/r**2) * (156*r6-42) * r6)

    def derivative(self, n=1):
        if n == 1:
            return self.first_derivative
        elif n == 2:
            return self.second_derivative
        else:
            raise ValueError(
                "Don't know how to compute {}-th derivative.".format(n))


###

class FeneLJCut():
    """
    Finite extensible nonlinear elastic(FENE) potential for a bead-spring polymer model.
    For the Lennard-Jones interaction a LJ-cut potential is used. Due to choice of the cutoff (rc=2^(1/6) sigma)
    it ensures a continous potential and force at the cutoff.
    """

    def __init__(self, K, R0, epsilon, sigma):
        self.K = K
        self.R0 = R0
        self.epsilon = epsilon
        self.sigma = sigma

    def __call__(self, r):
        """
        Return function value (potential energy).
        """
        r6 = (self.sigma/r)**6
        bond = -0.5 * self.K * self.R0**2 * np.log(1-(r/self.R0)**2)
        lj = 4 * self.epsilon * (r6-1) * r6 + self.epsilon
        return bond + lj

    def first_derivative(self, r):
        r6 = (self.sigma/r)**6
        bond = self.K * r / (1-(r/self.R0)**2)
        lj = -24 * self.epsilon * (2*r6/r-1/r) * r6
        return bond + lj

    def second_derivative(self, r):
        r6 = (self.sigma/r)**6
        invLength = 1 / (1-(r/self.R0)**2)
        bond = K * invLength + 2 * K * r**2 * invLength**2 / self.R0**2
        lj = 4 * self.epsilon * ((1/r**2) * (156*r6-42) * r6)
        return bond + lj

    def derivative(self, n=1):
        if n == 1:
            return self.first_derivative
        elif n == 2:
            return self.second_derivative
        else:
            raise ValueError(
                "Don't know how to compute {}-th derivative.".format(n))


###

class LennardJones84():
    """
    Function form of a 8-4 Lennard-Jones potential, used to model the structure of a CuZr.
    Kobayashi, Shinji et. al. "Computer simulation of atomic structure of Cu57Zr43 amorphous alloy."
    Journal of the Physical Society of Japan 48.4 (1980): 1147-1152.
    """

    def __init__(self, C1, C2, C3, C4, cutoff):
        self.C1 = C1
        self.C2 = C2
        self.C3 = C3
        self.C4 = C4
        self.cutoff = cutoff

    def get_cutoff():
        return self.cutoff

    def __call__(self, r):
        """
        Return function value (potential energy).
        """
        r4 = (1 / r)**4
        return (self.C2*r4-self.C1) * r4 + self.C3 * r + self.C4

    def first_derivative(self, r):
        r4 = (1 / r)**4
        return (-8 * self.C2*r4/r+4*self.C1/r) * r4 + self.C3

    def second_derivative(self, r):
        r4 = (1 / r)**4
        return (72 * self.C2 * r4 / r**2 - 20 * self.C1 / r**2) * r4

    def derivative(self, n=1):
        if n == 1:
            return self.first_derivative
        elif n == 2:
            return self.second_derivative
        else:
            raise ValueError(
                "Don't know how to compute {}-th derivative.".format(n))

###


class PairPotential(Calculator):
    implemented_properties = ['energy', 'stress', 'forces', "hessian"]
    default_parameters = {}
    name = 'PairPotential'

    def __init__(self, f, cutoff=None):
        Calculator.__init__(self)
        self.f = f

        self.dict = {x: obj.get_cutoff() for x, obj in f.items()}
        self.df = {x: obj.derivative(1) for x, obj in f.items()}
        self.df2 = {x: obj.derivative(2) for x, obj in f.items()}

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        nat = len(self.atoms)
        atnums = self.atoms.numbers
        atnums_in_system = set(atnums)

        i_n, j_n, dr_nc, abs_dr_n = neighbour_list(
            'ijDd', self.atoms, self.dict)

        e_n = np.zeros_like(abs_dr_n)
        de_n = np.zeros_like(abs_dr_n)
        for params, pair in enumerate(self.dict):
            if pair[0] == pair[1]:
                mask1 = atnums[i_n] == pair[0]
                mask2 = atnums[j_n] == pair[0]
                mask = np.logical_and(mask1, mask2)

                e_n[mask] = self.f[pair](abs_dr_n[mask])
                de_n[mask] = self.df[pair](abs_dr_n[mask])

            if pair[0] != pair[1]:
                mask1 = np.logical_and(
                    atnums[i_n] == pair[0], atnums[j_n] == pair[1])
                mask2 = np.logical_and(
                    atnums[i_n] == pair[1], atnums[j_n] == pair[0])
                mask = np.logical_or(mask1, mask2)

                e_n[mask] = self.f[pair](abs_dr_n[mask])
                de_n[mask] = self.df[pair](abs_dr_n[mask])

        epot = 0.5*np.sum(e_n)

        # Forces
        df_nc = -0.5*de_n.reshape(-1, 1)*dr_nc/abs_dr_n.reshape(-1, 1)

        # Sum for each atom
        fx_i = np.bincount(j_n, weights=df_nc[:, 0], minlength=nat) - \
            np.bincount(i_n, weights=df_nc[:, 0], minlength=nat)
        fy_i = np.bincount(j_n, weights=df_nc[:, 1], minlength=nat) - \
            np.bincount(i_n, weights=df_nc[:, 1], minlength=nat)
        fz_i = np.bincount(j_n, weights=df_nc[:, 2], minlength=nat) - \
            np.bincount(i_n, weights=df_nc[:, 2], minlength=nat)

        # Virial
        virial_v = -np.array([dr_nc[:, 0]*df_nc[:, 0],               # xx
                              dr_nc[:, 1]*df_nc[:, 1],               # yy
                              dr_nc[:, 2]*df_nc[:, 2],               # zz
                              dr_nc[:, 1]*df_nc[:, 2],               # yz
                              dr_nc[:, 0]*df_nc[:, 2],               # xz
                              dr_nc[:, 0]*df_nc[:, 1]]).sum(axis=1)  # xy

        self.results = {'energy': epot,
                        'stress': virial_v/self.atoms.get_volume(),
                        'forces': np.transpose([fx_i, fy_i, fz_i])}

    ###

    def calculate_hessian_matrix(self, atoms, H_format="dense", limits=None):
        """
        Calculate the Hessian matrix for a pair potential.
        For an atomic configuration with N atoms in d dimensions the hessian matrix is a symmetric, hermitian matrix
        with a shape of (d*N,d*N). The matrix is in general a sparse matrix, which consists of dense blocks of shape (d,d), which
        are the mixed second derivatives. The result of the derivation for a pair potential can be found in:
        L. Pastewka et. al. "Seamless elastic boundaries for atomistic calculations", Phys. Ev. B 86, 075459 (2012).

        Parameters
        ----------
        atoms: ase.Atoms
            Atomic configuration in a local or global minima.

        H_format: "dense" or "sparse"
            Output format of the hessian matrix.
            The format "sparse" is only possible if matscipy was build with scipy.

        limits: list [atomID_low, atomID_up]
            Calculate the Hessian matrix only for the given atom IDs. 
            If limits=[5,10] the Hessian matrix is computed for atom IDs 5,6,7,8,9 only.
            The Hessian matrix will have the full shape dim(3*N,3*N) where N is the number of atoms. 
            This ensures correct indexing of the data. 

        Restrictions
        ----------
        This method is currently only implemented for three dimensional systems

        """

        if H_format == "sparse":
            try:
                from scipy.sparse import bsr_matrix, vstack, hstack
            except ImportError:
                raise ImportError(
                    "Import error: Can not output the hessian matrix since scipy.sparse could not be loaded!")

        f = self.f
        dict = self.dict
        df = self.df
        df2 = self.df2

        nat = len(atoms)
        atnums = atoms.numbers

        i_n, j_n, dr_nc, abs_dr_n = neighbour_list('ijDd', atoms, dict)
        first_i = first_neighbours(nat, i_n)

        e_n = np.zeros_like(abs_dr_n)
        de_n = np.zeros_like(abs_dr_n)
        dde_n = np.zeros_like(abs_dr_n)
        for params, pair in enumerate(dict):
            if pair[0] == pair[1]:
                mask1 = atnums[i_n] == pair[0]
                mask2 = atnums[j_n] == pair[0]
                mask = np.logical_and(mask1, mask2)

                e_n[mask] = f[pair](abs_dr_n[mask])
                de_n[mask] = df[pair](abs_dr_n[mask])
                dde_n[mask] = df2[pair](abs_dr_n[mask])

            if pair[0] != pair[1]:
                mask1 = np.logical_and(
                    atnums[i_n] == pair[0], atnums[j_n] == pair[1])
                mask2 = np.logical_and(
                    atnums[i_n] == pair[1], atnums[j_n] == pair[0])
                mask = np.logical_or(mask1, mask2)

                e_n[mask] = f[pair](abs_dr_n[mask])
                de_n[mask] = df[pair](abs_dr_n[mask])
                dde_n[mask] = df2[pair](abs_dr_n[mask])

        if limits != None:
            if limits[1] < limits[0]:
                raise ValueError(
                    "Value error: The upper atom id cannot be smaller than the lower atom id.")
            else:
                mask = np.logical_and(i_n >= limits[0], i_n < limits[1])
                i_n = i_n[mask]
                i_n1 = i_n - i_n[0]
                j_n = j_n[mask]
                dr_nc = dr_nc[mask]
                abs_dr_n = abs_dr_n[mask]
                e_n = e_n[mask]
                de_n = de_n[mask]
                dde_n = dde_n[mask]
                nat1 = limits[1] - limits[0]

                first_i = [0] * (nat1 + 1)
                j = 1
                for k in range(1, len(i_n)):
                    if i_n[k] != i_n[k-1]:
                        first_i[j] = k
                        j = j+1
                first_i[-1] = len(i_n)

                if H_format == "sparse":
                    # Off-diagonal elements of the Hessian matrix
                    e_nc = (dr_nc.T / abs_dr_n).T
                    H_ncc = -(dde_n * (e_nc.reshape(-1, 3, 1)
                                       * e_nc.reshape(-1, 1, 3)).T).T
                    H_ncc += -(de_n / abs_dr_n * (np.eye(3, dtype=e_nc.dtype)
                                                  - (e_nc.reshape(-1, 3, 1) * e_nc.reshape(-1, 1, 3))).T).T

                    H_nat1nat = bsr_matrix(
                        (H_ncc, j_n, first_i), shape=(3*nat1, 3*nat))

                    # Stack matrices in order to obtain full shape (3*nat, 3*nat)
                    H = vstack([bsr_matrix((limits[0]*3, 3*nat)), H_nat1nat,
                                bsr_matrix((3*nat - limits[1]*3, 3*nat))])

                    # Diagonal elements of the Hessian matrix
                    Hdiag_icc = np.empty((nat1, 3, 3))
                    for x in range(3):
                        for y in range(3):
                            Hdiag_icc[:, x, y] = - \
                                np.bincount(i_n1, weights=H_ncc[:, x, y])

                    Hdiag_nat1nat = bsr_matrix((Hdiag_icc, np.arange(limits[0], limits[1]),
                                                np.arange(nat1+1)), shape=(3*nat1, 3*nat))

                    # Compute full Hessian matrix
                    H += vstack([bsr_matrix((limits[0]*3, 3*nat)), Hdiag_nat1nat,
                                 bsr_matrix((3*nat - limits[1]*3, 3*nat))])

                    return H

                elif H_format == "dense":
                    # Off-diagonal elements of the Hessian matrix
                    e_nc = (dr_nc.T / abs_dr_n).T
                    H_ncc = -(dde_n * (e_nc.reshape(-1, 3, 1) *
                                       e_nc.reshape(-1, 1, 3)).T).T
                    H_ncc += -(de_n/abs_dr_n * (np.eye(3, dtype=e_nc.dtype)
                                                - (e_nc.reshape(-1, 3, 1) * e_nc.reshape(-1, 1, 3))).T).T

                    H = np.zeros((3*nat, 3*nat))
                    for atom in range(len(i_n)):
                        H[3*i_n[atom]:3*i_n[atom]+3,
                          3*j_n[atom]:3*j_n[atom]+3] += H_ncc[atom]

                    # Diagonal elements of the Hessian matrix
                    Hdiag_icc = np.empty((nat1, 3, 3))
                    for x in range(3):
                        for y in range(3):
                            Hdiag_icc[:, x, y] = - \
                                np.bincount(i_n1, weights=H_ncc[:, x, y])

                    Hdiag_ncc = np.zeros((3*nat, 3*nat))
                    for atom in range(nat1):
                        Hdiag_ncc[3*(atom+limits[0]):3*(atom+limits[0])+3,
                                  3*(atom+limits[0]):3*(atom+limits[0])+3] += Hdiag_icc[atom]

                    # Compute full Hessian matrix
                    H += Hdiag_ncc

                    return H

        # Sparse BSR-matrix
        elif H_format == "sparse":
            e_nc = (dr_nc.T/abs_dr_n).T
            H_ncc = -(dde_n * (e_nc.reshape(-1, 3, 1)
                               * e_nc.reshape(-1, 1, 3)).T).T
            H_ncc += -(de_n/abs_dr_n * (np.eye(3, dtype=e_nc.dtype)
                                        - (e_nc.reshape(-1, 3, 1) * e_nc.reshape(-1, 1, 3))).T).T

            H = bsr_matrix((H_ncc, j_n, first_i), shape=(3*nat, 3*nat))

            Hdiag_icc = np.empty((nat, 3, 3))
            for x in range(3):
                for y in range(3):
                    Hdiag_icc[:, x, y] = - \
                        np.bincount(i_n, weights=H_ncc[:, x, y])

            H += bsr_matrix((Hdiag_icc, np.arange(nat),
                             np.arange(nat+1)), shape=(3*nat, 3*nat))
            return H

        # Dense matrix format
        elif H_format == "dense":
            e_nc = (dr_nc.T/abs_dr_n).T
            H_ncc = -(dde_n * (e_nc.reshape(-1, 3, 1)
                               * e_nc.reshape(-1, 1, 3)).T).T
            H_ncc += -(de_n/abs_dr_n * (np.eye(3, dtype=e_nc.dtype)
                                        - (e_nc.reshape(-1, 3, 1) * e_nc.reshape(-1, 1, 3))).T).T

            H = np.zeros((3*nat, 3*nat))
            for atom in range(len(i_n)):
                H[3*i_n[atom]:3*i_n[atom]+3,
                  3*j_n[atom]:3*j_n[atom]+3] += H_ncc[atom]

            Hdiag_icc = np.empty((nat, 3, 3))
            for x in range(3):
                for y in range(3):
                    Hdiag_icc[:, x, y] = - \
                        np.bincount(i_n, weights=H_ncc[:, x, y])

            Hdiag_ncc = np.zeros((3*nat, 3*nat))
            for atom in range(nat):
                Hdiag_ncc[3*atom:3*atom+3,
                          3*atom:3*atom+3] += Hdiag_icc[atom]

            H += Hdiag_ncc

            return H



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

"""
Interface from ASE to the chemview Jupyter visualiser.

Your Jupyter notebook will need to contain

from chemview import enable_notebook
enable_notebook()
"""

###

import itertools

import numpy as np

from ase.data import covalent_radii
from matscipy.neighbours import neighbour_list

from chemview import MolecularViewer

###

def view(a, colour=None, bonds=True, cell=True,
         scale=10.0, cutoff_scale=1.2,
         cmap=None, vmin=None, vmax=None):
    topology = {}
    topology['atom_types'] = a.get_chemical_symbols()

    if bonds:
        n = a.numbers
        maxn = n.max()
        cutoffs = np.zeros([maxn+1, maxn+1])

        for n1, n2 in itertools.product(n, n):
            cutoffs[n1, n2] = cutoff_scale*(covalent_radii[n1]+covalent_radii[n2])

        # Construct a bond list
        i, j, S = neighbour_list('ijS',
                                 a, cutoffs,
                                 np.array(a.numbers, dtype=np.int32))
        m = np.logical_and(i<j, (S==0).all(axis=1))
        i = i[m]
        j = j[m]
        topology['bonds'] = [(x, y) for x, y in zip(i, j)]

    colorlist = None
    if colour is not None:
        colour = np.array(colour, dtype=np.float64)
        if cmap is None:
            from matplotlib.cm import jet
            cmap = jet
        if vmin is None:
            vmin = np.min(colour)
        if vmax is None:
            vmax = np.max(colour)
        colour = (colour - vmin)/(vmax - vmin)
        colorlist = ['0x%02x%02x%02x' % (r*256, g*256, b*256)
                     for (r, g, b, alpha) in cmap(colour)]

    mv = MolecularViewer(a.positions/scale,
                         topology=topology)
    mv.ball_and_sticks(colorlist=colorlist)

    if cell:
        O = np.zeros(3, dtype=np.float32)
        La, Lb, Lc = a.cell.astype(np.float32)/scale
        start = np.r_[O, O, O,
                      O + Lb, O + Lc, O + La,
                      O + Lc, O + La, O + Lb,
                      O + Lb + Lc, O + La + Lc, O + La + Lb]
        end = np.r_[O + La, O + Lb, O + Lc,
                    O + Lb + La, O + Lc + Lb, O + La + Lc,
                    O + Lc + La, O + La + Lb, O + Lb + Lc,
                    O + Lb + Lc + La, O + La + Lc + Lb, O + La + Lb + Lc]
        rgb = [0xFF0000, 0x00FF00, 0x0000FF]*4
        mv.add_representation('lines', {'startCoords': start,
                                        'endCoords': end,
                                        'startColors': rgb,
                                        'endColors': rgb})
    return mv



from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np

from .base_qm_cluster_tool import BaseQMClusterTool


class QMFlaggingTool(BaseQMClusterTool):
    """This class is responsible for flagging atoms
    that move out of their equilibrium"""

    def __init__(self, mediator=None, qm_flag_potential_energies=None,
                 small_cluster_hops=3, only_heavy=False, ema_parameter=0.1, energy_cap=None,
                 energy_increase=1):
        """This class is responsible for flagging atoms
        that move out of their equilibrium

        Parameters
        ----------
        mediator : matscipy.calculators.mcfm.QMCluster
            class responsible for managing the QM clusters in the simulation
        qm_flag_potential_energies : np.array
            threshholds for flagging indivual atoms.
            The diensions are (nAtoms, 2) where:
                column 1: threshold to enter the QM regios
                column 2: threshold to stay the QM region
        small_cluster_hops : int
            Each flagged atom and atoms around it within small_cluster_hops neighbour hops
            will generate a single cluster, clusters are later joined.
        only_heavy : bool
            If True, only consider non-hydrogen atoms in cluster expansion.
            Hydrogens are added later
        ema_parameter : float
            parameter lambda in the exponential mean average calculation
        energy_cap : float
            if not None, cap potential energy per atom at this value
        energy_increase : int
            Multiplier for potential energy per atom, used to scale it for convininece

        """
        # Initialize the QMClusterObject with a mediator
        super(QMFlaggingTool, self).__init__(mediator)

        try:
            self.qm_flag_potential_energies = qm_flag_potential_energies
        except AttributeError:
            raise AttributeError("QM flag PE/force tolerance must be defined")

        self.small_cluster_hops = small_cluster_hops
        self.only_heavy = only_heavy
        self.ema_parameter = ema_parameter
        self.energy_cap = energy_cap
        self.energy_increase = energy_increase

        self.qm_atoms_list = []
        self.old_energized_list = []

        self.verbose = 0

    def get_energized_list(self, atoms, data_array, property_str, hysteretic_tolerance):
        """Produce a list of atoms that are ot be flagged as a QM region
        based on the properties given in the array according to the
        tolerance given.

        Parameters
        ----------
        atoms : ase.Atoms
            Whole structure
        data_array : array
            an array of per atom data providing information
        property_str : str
            name of th property so that it can be stored in atoms.properties.
        hysteretic_tolerance : array
            Threshholds for flagging indivual atoms.
            The diensions are (nAtoms, 2) where:
                column 1: threshold to enter the QM regios
                column 2: threshold to stay the QM region

        Returns
        -------
        list
            List of flagged atoms
        """

        # ------ Update EPA
        update_avg_property_per_atom(atoms, data_array, property_str, self.ema_parameter)

        avg_property_per_atom = atoms.arrays[property_str]

        tolerance = np.zeros(len(atoms)) + hysteretic_tolerance[:, 0]
        tolerance[self.old_energized_list] = hysteretic_tolerance[self.old_energized_list, 1]
        energized_mask = np.greater_equal(avg_property_per_atom, tolerance)
        energized_list = np.arange(len(atoms))[energized_mask]

        return energized_list

    def create_cluster_around_atom(self, atoms, atom_id, hydrogenate=False):
        """Carve a cluster around the atom with atom_id
        This function operates on sets and returns a set

        Parameters
        ----------
        atoms : ase.Atoms
            Whole structure
        atom_id : int
            Atomic index
        hydrogenate : bool
            If true, hydrogenate the resulting structure

        Returns
        -------
        list
            atoms in the new cluster
        """
        cluster_set = set([atom_id])
        edge_neighbours = set([atom_id])

        for i in range(self.small_cluster_hops):
            new_neighbours = set()
            # For each atom in edge neighbours list, expand the list
            for index in edge_neighbours:
                new_neighbours |= set(self.find_neighbours(atoms, index)[0])
            # Remove atoms already in the qm list
            edge_neighbours = new_neighbours - cluster_set
            # Make a union of the sets
            cluster_set = cluster_set | edge_neighbours

        # ----- If specified, add hydrogens ot the cluster
        if hydrogenate:
            self.hydrogenate_cluster(atoms, cluster_set)
        return cluster_set

    def join_clusters(self, verbose=False):
        """This function will join the clusters if they overlap
        Input is an array of sets each representing individual
        small cluster

        Parameters
        ----------
        verbose : bool
            Print messages during calculation
        """

        i = 0
        # Iterate over the whole list C taking into account that it might get
        # throughout the loop
        while (i < len(self.qm_atoms_list)):

            # Iterate over the sets taking into account that C can change
            # Do not repeat pairise disjointment checks
            # i.e. for a list of sets [A, B, C, D]
            # first loop included checks A-B, A-C, A-D (pairs 0 - 1:3)
            # Then make sure the second only does B-C, B-D (pairs 1 - 2:3)
            for j in range(i + 1, len(self.qm_atoms_list)):
                if verbose is True:
                    print(i, j, self.qm_atoms_list[i], self.qm_atoms_list[j],
                          not set.isdisjoint(self.qm_atoms_list[i], self.qm_atoms_list[j]))

                if not set.isdisjoint(self.qm_atoms_list[i], self.qm_atoms_list[j]):
                    # If intersection detected, unify sets
                    self.qm_atoms_list[i] |= self.qm_atoms_list[j]
                    # Then delete the second set to avoid duplicates
                    # Then restart the j loop to see if now, any set
                    # has an intersection with the new union
                    del self.qm_atoms_list[j]
                    i -= 1

                    if verbose is True:
                        for entry in self.qm_atoms_list:
                            print(entry)
                    break
            i += 1

    def expand_cluster(self, special_atoms_list):
        """Include extra atoms in the cluster.

        If one of the special atoms is included in one of the clusters,
        add all other special atoms to this cluster

        Parameters
        ----------
        special_atoms_list : list
            list of the special atoms
        """

        for specialMolecule in special_atoms_list:
            specialMoleculeSet = set(specialMolecule)
            for clusterIndex in range(len(self.qm_atoms_list)):
                if (not specialMoleculeSet.isdisjoint(self.qm_atoms_list[clusterIndex])):
                    self.qm_atoms_list[clusterIndex] |= specialMoleculeSet

    def update_qm_region(self, atoms,
                         potential_energies=None,
                         ):
        """Update the QM region while the simulation is running

        Parameters
        ----------
        atoms : ase.Atoms
            whole structure
        potential_energies : array
            Potential energy per atom

        Returns
        -------
        list of lists of ints
            list of individual clusters as lists of atoms
        """
        # Make sure the right atoms object is in

        # ------ Increase the energy by a common factor - makes it more readable in some cases
        if (self.energy_increase is not None):
            potential_energies *= self.energy_increase

        # ------ Cap maximum energy according to the flag
        if (self.energy_cap is not None):
            np.minimum(potential_energies, self.energy_cap, potential_energies)

        # ------ Get the energized atoms list
        flagged_atoms_dict = {}

        flagged_atoms_dict["potential_energies"] = self.get_energized_list(atoms,
                                                                           potential_energies,
                                                                           "avg_potential_energies",
                                                                           self.qm_flag_potential_energies)

        energized_set = set()
        for key in flagged_atoms_dict:
            energized_set = set(flagged_atoms_dict[key]) | energized_set
        energized_list = list(energized_set)
        self.old_energized_list = list(energized_list)

        if (len(energized_list) != 0):
            self.mediator.neighbour_list.update(atoms)

        # TODO if energized list include the whole system just pass it along
        for array_i, atom_i in enumerate(energized_list):
            energized_list[array_i] = self.create_cluster_around_atom(atoms, atom_i, hydrogenate=False)

        self.qm_atoms_list = energized_list
        if (len(self.qm_atoms_list) > 0):
            self.join_clusters()
            self.expand_cluster(self.mediator.special_atoms_list)
            self.join_clusters()

        if self.only_heavy is False:
            for index in range(len(self.qm_atoms_list)):
                self.qm_atoms_list[index] = self.hydrogenate_cluster(atoms, self.qm_atoms_list[index])

        self.qm_atoms_list = list(map(list, self.qm_atoms_list))
        return self.qm_atoms_list
        # print "QM cluster", self.qm_atoms_list


def exponential_moving_average(oldset, newset=None, ema_parameter=0.1):
    """Apply the exponential moving average to the given array

    Parameters
    ----------
    oldset : array
        old values
    newset : array
        new data set
    ema_parameter : float
        parameter lambda
    """
    if newset is None:
        pass
    else:
        oldset *= (1 - ema_parameter)
        oldset += ema_parameter * newset


def update_avg_property_per_atom(atoms, data_array, property_str, ema_parameter):
    """Update the per atom property using running avarages
    and store it in atoms.properties[property_str]

    Parameters
    ----------
    atoms : ase.Atoms
        structure that need updated values
    data_array : array
        data that need to be attached to atoms
    property_str : str
        key for structure properties dictionary
    ema_parameter : float
        Coefficient for the Exponential Moving Average
    """

    # Abbreviations
    # ppa - (property per atom
    # appa - average property per atom

    ppa = data_array

    # ------ Get average ppa
    if (property_str in atoms.arrays):
        exponential_moving_average(atoms.arrays[property_str],
                                   ppa, ema_parameter)
    else:
        atoms.arrays[property_str] = ppa.copy()



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

from math import pi

import numpy as np

###

def J_integral(a, deformation_gradient, virial, epot, e0, tip_x, tip_y, r1, r2,
               mask=None):
    """
    Compute the energy release rate from the J-integral. Converts contour
    integral into a domain integral.
    See: Li, Shih, Needleman, Eng. Fract. Mech. 21, 405 (1985);
    Jin, Yuan, J. Nanosci. Nanotech. 5, 2099 (2005)
    Domain function is currently fixed: q(r) = (r-r1)/(r2-r1)

    Parameters
    ----------
    a : ase.Atoms
        Relaxed atomic configuration of the crack.
    deformation_gradient : array_like
        len(a) x 3x3 array of atomic deformation gradients.
    virial : array_like
        len(a) x 3x3 array of atomic virials.
    epot : float
        Potential energy per atom of the cracked configuration.
    e0 : float
        Reference energy (cohesive energy per atom of the bulk crystal at
        equilibrium).
    tip_x, tip_y : float
        Position of the crack tip.
    r1, r2 : float
        Volume integration is carried out in region at a distance between r1
        and r2 from the crack tip.
    mask : array_like
        Include only a subset of all atoms into J-integral computation.

    Returns
    -------
    J : float
        Value of the J-integral.
    """

    if mask is None:
        mask = np.ones(len(a), dtype=bool)

    # Cell size
    sx, sy, sz = a.cell.diagonal()

    # Positions
    x, y, z = a.positions.T.copy()
    x -= tip_x
    y -= tip_y
    r = np.sqrt(x**2+y**2)

    # Derivative of the domain function q
    nonzero = np.logical_and(r > r1, r < r2)
    # q = (r-r1)/(r2-r1)
    if 0:
        gradq = np.transpose([
                np.where(nonzero,
                         x/((r2-r1)*r),
                         np.zeros_like(x)),
                np.where(nonzero,
                         y/((r2-r1)*r),
                         np.zeros_like(y)),
                np.zeros_like(z)])
    elif 1:
    # q = (1-cos(pi*(r-r1)/(r2-r1)))/2
        gradq = np.transpose([
                np.where(nonzero,
                         x/r * pi/(2*(r2-r1)) * np.sin(pi*(r-r1)/(r2-r1)),
                         np.zeros_like(x)),
                np.where(nonzero,
                         y/r * pi/(2*(r2-r1)) * np.sin(pi*(r-r1)/(r2-r1)),
                         np.zeros_like(y)),
                np.zeros_like(z)])
    else:
    # q = (2*pi*(r-r1) - (r2-r1)*sin(2*pi*(r-r1)/(r2-r1))) / (2*pi*(r2-r1))
    # dq/dq = (1 - cos(2*pi*(r-r1)/(r2-r1))) / (r2-r1)
        gradq = np.transpose([
                np.where(nonzero,
                         x/r * (1. - np.cos(2*pi*(r-r1)/(r2-r1))) / (r2-r1),
                         np.zeros_like(x)),
                np.where(nonzero,
                         y/r * (1. - np.cos(2*pi*(r-r1)/(r2-r1))) / (r2-r1),
                         np.zeros_like(y)),
                np.zeros_like(z)])

    # Potential energy
    Jpot = ((epot[mask]-e0[mask])*gradq[mask,0]).sum()

    # Strain energy
    Jstrain = np.einsum('aij,ai,aj->', virial[mask,:,:],
                        deformation_gradient[mask,:,0], gradq[mask,:])

    # Compute J-integral
    return (Jpot-Jstrain)/sz




# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#                  Adrien Gola, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

"""
Helper routines for potential fitting
"""

from __future__ import print_function

from math import atanh, sqrt, tanh, isnan
import sys

import numpy as np

import ase
import ase.constraints
import ase.io
import ase.lattice.compounds as compounds
import ase.lattice.cubic as cubic
import ase.lattice.hexagonal as hexagonal
import ase.optimize

from ase.units import GPa,J,m

import scipy
scipy_v = scipy.__version__

if int(scipy_v.split('.')[1]) <= 14 : 
    from scipy.optimize import minimize, leastsq, anneal, brute
else :
    # scipy.optimize.anneal decprecated from version 0.14.0, documentation advise to use scipy.optimize.basinhopping instead
    from scipy.optimize import minimize, leastsq, brute
    from scipy.signal import argrelextrema
    

try:
    from openopt import GLP
    have_openopt = True
except:
    have_openopt = False

###

_logfile = None
Jm2 = 1e23/ase.units.kJ

### Parameters

class Parameters(object):
    """
    Stores a parameter set for fitting purposes.

    In particular, it distinguishes between variable parameters (to be fitted),
    constant parameter and derived parameters, the latter ones having a
    functional dependence on the other parameters.
    """

    __slots__ = [ 'default', 'variable', 'constant', 'derived', 'hidden', 
                  'parameters', 'ranges', 'range_mapping' ]

    def __init__(self, default, constant, derived, ranges={}, hidden=[]):
        """
        Parameters
        ----------
        default : dict
            Dictionary with the default parameter set
        constant : list
            List of parameters that are constant
        derived : dict
            Dictionary with derived parameters and a function to get those
            derived values.
        """
        self.default = default
        self.constant = set(constant)
        self.derived = derived
        self.ranges = ranges
        self.hidden = set(hidden)

        self.parameters = self.default.copy()

        self.range_mapping = False

        self._update_variable()
        self._update_derived()

    def set_range_derived(self):
        self.range_mapping = True
        for key, ( x1, x2 ) in self.ranges.items():
            if x1 > x2:
                raise RuntimeError('Inverted ranges {0}:{1} for parameter {2}.'
                                   .format(x1, x2, key))
            x = self.parameters[key]
            if x <= x1 or x >= x2:
                raise ValueError('Parameter {0} has value {1} which is '
                                 'outside of the bounds {2}:{3}.'
                                 .format(key, x, x1, x2))
            self.parameters[':'+key] = atanh(2*(x-x1)/(x2-x1)-1)

            self.variable += [ ':'+key ]
            self.hidden.add(':'+key)

    def _update_variable(self):
        self.variable = []
        for key in self.default.keys():
            if not ( key in self.constant or key in self.derived or 
                     key in self.ranges ):
                self.variable += [ key ]

    def _update_constant(self):
        self.constant = []
        for key in self.default.keys():
            if not ( key in self.variable or key in self.derived ):
                self.constant += [ key ]

    def _update_derived(self):
        if self.range_mapping:
            for key, ( x1, x2 ) in self.ranges.items():
                self.parameters[key] = x1+\
                    0.5*(x2-x1)*(1+tanh(self.parameters[':'+key]))
        for key, func in self.derived.items():
            self.parameters[key] = func(self.parameters)

    def __len__(self):
        return len(self.variable)

    def set_variable(self, variable):
        self.variable = variable
        self._update_constant()

    def get_variable(self):
        return self.variable

    def set_constant(self, constant):
        self.constant = constant
        self._update_variable()

    def get_constant(self):
        return self.constant

    def set_derived(self, derived):
        self.derived = derived
        self._update_variable()
        self._update_derived()

    def get_derived(self):
        return self.derived

    def set(self, key, value):
        if key in self.derived:
            raise RuntimeError('Cannot set parameter %s since it is a derived '
                               'quantity.' % key)

        self.parameters[key] = value

        self._update_derived()
    __setitem__ = set

    def get(self, key):
        return self.parameters[key]
    __getitem__ = get

    def __getattr__(self, key):
        try:
            return self.get(key)
        except:
            return super(Parameters, self).__getattr__(key)

    def set_dict(self, d):
        for key, value in d.items():
            self.set(key, value)

    def get_dict(self):
        p = self.parameters.copy()
        if self.hidden is not None:
            for key in self.hidden:
                del p[key]
        return p

    def set_array(self, a, keys=None):
        if keys is None:
            keys = self.variable

        for key, value in zip(keys, a):
            self.set(key, value)

    def get_array(self, keys=None):
        if keys is None:
            keys = self.variable

        return [ self.parameters[key] for key in keys ]

    def get_lower_bounds(self, keys=None):
        if keys is None:
            keys = self.variable
        return [ self.ranges[key][0] for key in keys ]

    def get_upper_bounds(self, keys=None):
        if keys is None:
            keys = self.variable
        return [ self.ranges[key][1] for key in keys ]

    def in_range(self, key=None):
        if self.ranges is None:
            return True
        if not key is None:
            if key in self.ranges:
                x1, x2 = self.ranges[key]
                return self.parameters[key] >= x1 and self.parameters[key] <= x2
            else:
                return True

        r = True
        for key, value in self.parameters.items():
            if key in self.ranges:
                x1, x2 = self.ranges[key]
                r = r and self.parameters[key] >= x1 and \
                    self.parameters[key] <= x2
        return r

    def __str__(self):
        s = ''
        for key, value in self.parameters.items():
            if not key.startswith(':'):
                s += '# {0:>24s} = {1}\n'.format(key, value)
        return s

### Fitting superclass

class Fit(object):
    """
    Parameter optimization class.
    """

    __slots__ = [ 'atoms', 'calc', 'cost_history', 'minimal_cost', 'par',
                  'par_at_minimal_cost', 'residuals_history' ]

    def __init__(self, calc, par):
        self.calc = calc
        self.par = par

        self.minimal_cost = 1e40
        self.par_at_minimal_cost = None

        self.cost_history = []
        self.residuals_history = []

    def set_parameters_from_array(self, p):
        self.par.set_array(p)
        self.set_calculator(self.calc(**self.par.get_dict()))

    def set_calculator_class(self, calc):
        self.calc = calc

    def set_parameters(self, par):
        self.par = par
        self.set_calculator(self.calc(**self.par.get_dict()))

    def get_potential_energy(self):
        return self.atoms.get_potential_energy()

    def get_cohesive_energy(self):
        return self.atoms.get_potential_energy()/len(self.atoms)

    def get_square_residuals(self, p=None, log=None):
        r2 = 0.0
        if p is not None:
            self.set_parameters_from_array(p)
        r = np.array(self.get_residuals(log=log))
        self.residuals_history += [ r ]
        return r*r

    def get_cost_function(self, p=None, log=None, store_history=False):
        try:
            c = np.sum(self.get_square_residuals(p, log=log))
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            print('# Warning: Error computing residuals. Penalizing '
                  'parameters but continuing. Error message: {0}' \
                  .format(e), file=log)
            c = 1e40
        if log is not None:
            print('# Current value of the cost function/residual: {0}' \
                .format(c), file=log)
        if isnan(c):
            c = 1e40
        if store_history:
            if self.cost_history == [] or c < np.min(self.cost_history):
                print('# New minimum of cost function: {0}'.format(c))
                self.minimal_cost = c
                self.par_at_minimal_cost = p
            self.cost_history += [c]
        return c

    def get_residuals_history(self):
        return np.array(self.residuals_history)

    def get_cost_history(self):
        return np.array(self.cost_history)

    def get_parameters_at_minimal_cost(self, log=sys.stdout):
        self.set_parameters_from_array(self.par_at_minimal_cost)
        print('=== PARAMETERS AT MINIMAL COST ===', file=log)
        self.get_cost_function(self.par_at_minimal_cost, log=log)
        return self.par

    def optimize(self, log=sys.stdout, **kwargs):
        self.par.set_range_derived()
        res = minimize(self.get_cost_function, self.par.get_array(),
                       args=(log, True,), **kwargs)
        self.set_parameters_from_array(res.x)
        print('=== HISTORY OF COST FUNCTION ===', file=log)
        print(self.cost_history, file=log)
        print('=== FINAL OPTIMIZED PARAMETER SET ===', file=log)
        final_cost = self.get_cost_function(res.x, log)
        if abs(final_cost - self.minimal_cost) > 1e-6:
            print('# WARNING: Final cost (={0}) is not minimal (={1}). This '
                  'may come from insufficiently converged calculations. Try '
                  'to decrease fmax.'.format(final_cost, self.minimal_cost))
        return self.par

    def optimize_leastsq(self, log=sys.stdout):
        self.par.set_range_derived()
        self.set_parameters_from_array(leastsq(self.get_square_residuals,
                                               self.par.get_array(),
                                               args=(log,))[0])
        return self.par

    def optimize_anneal(self, **kwargs):
        self.set_parameters_from_array(anneal(self.get_cost_function,
                                              self.par.get_array(),
                                              lower=self.par.get_lower_bounds(),
                                              upper=self.par.get_upper_bounds(),
                                              **kwargs))
        return self.par
      

    def optimize_brute(self, **kwargs):
        x0, fval = brute(
            self.get_cost_function,
            ranges=map(tuple,
                       list(np.transpose([self.par.get_lower_bounds(),
                                          self.par.get_upper_bounds()]))),
            **kwargs)
        self.set_parameters_from_array(x0)
        print('=== OPTIMIZED PARAMETER SET ===', file=log)
        self.get_cost_function(x0)
        return self.par

    def optimize_openopt(self, solver='interalg'):
        if not have_openopt:
            raise RuntimeError('OpenOpt not available.')
        p = GLP(self.get_cost_function, self.par.get_array(),
                lb=self.par.get_lower_bounds(),
                ub=self.par.get_upper_bounds())
        r = p.solve(solver)
        print(r, file=log)
        self.set_parameters_from_array(r.xf)
        print('=== OPTIMIZED PARAMETER SET ===', file=log)
        self.get_cost_function(r.xf)
        return self.par

class CombinedFit(Fit):

    __slots__ = [ 'targets' ]

    def __init__(self, calc, par, targets):
        Fit.__init__(self, calc, par)

        self.targets = targets

    def set_parameters_from_array(self, p):
        for target in self.targets:
            target.set_parameters_from_array(p)

    def set_parameters(self, p):
        for target in self.targets:
            target.set_parameters(p)

    def set_calculator_class(self, calc):
        for target in self.targets:
            target.set_calculator_class(calc)

    def get_residuals(self, log=None):
        if log is not None:
            print('', file=log)
            print('# Computing properties for parameter set:', file=log)
            print(self.par, file=log)
            f = open('par.out', 'w')
            print(self.par, file=f)
            f.close()
        r = []
        for target in self.targets:
            r = np.append(r, target.get_residuals(log=log))
        return r

    def get_potential_energy(self):
        raise RuntimeError('get_potential_energy does not make sense for the '
                           'CombinedFit class.')

class RotatingFit(object):

    __slots__ = [ 'targets', 'par' ]

    def __init__(self, par, targets):
        self.par = par

        self.targets = targets


    def optimize(self, pmax=1e-3, mix=None, **kwargs):
        globalv = np.array(self.par.get_variable()).copy()

        dp = 1e6
        while dp > pmax:
            p = np.array(self.par.get_array())

            for target, variable in self.targets:
                if log is not None:
                    print('# ===', target, '===', variable, '===', file=log)
                self.par.set_variable(variable)
                target.set_parameters(self.par)
                self.par = target.optimize(**kwargs)

            self.par.set_variable(globalv)

            cp = np.array(self.par.get_array())
            dp = cp - p
            dp = sqrt(np.sum(dp*dp))

            if mix is not None:
                self.par.set_array(mix*cp + (1-mix)*p)

        self.par.set_variable(globalv)

        return self.par


### Generic penalty function

class Penalty(Fit):

    __slots__ = [ 'func' ]

    def __init__(self, calc, par, func): 
        Fit.__init__(self, calc, par)
        self.func = func


    def set_calculator(self, calc):
        """
        Set the calculator
        """
        pass


    def get_residuals(self, log=None):
        return self.func(self.par, log=log)


### Single point

class FitSinglePoint(Fit):

    __slots__ = [ 'atoms', 'energy', 'forces', 'stress', 'w_energy', 'w_forces',
                  'w_stress' ]

    def __init__(self, calc, par, atoms, w_energy=None, w_forces=None,
                 w_stress=None): 
        Fit.__init__(self, calc, par)
        self.original_atoms = atoms
        self.w_energy = w_energy
        self.w_forces = w_forces
        self.w_stress = w_stress

        self.energy = self.atoms.get_potential_energy()
        self.forces = self.atoms.get_forces().copy()
        self.stress = self.atoms.get_stress().copy()

        self.atoms = None

    def set_calculator(self, calc):
        """
        Set the calculator
        """
        self.atoms = self.original_atoms.copy()
        self.atoms.set_calculator(calc)
        self.atoms.get_potential_energy()

    def get_residuals(self, log=None):
        r = []
        w = []
        if self.w_energy is not None:
            cr = [ self.w_energy*(
                    self.atoms.get_potential_energy()-self.energy
                    )/self.energy ]
            r += cr
            w += [ self.w_energy ]
        if self.w_forces is not None:
            cr = list(
                (self.w_forces*(
                        self.atoms.get_forces() - self.forces
                        )/self.forces).flatten()
                )
            r += cr
            w += list(self.w_forces*np.ones_like(self.forces).flatten())
        if self.w_stress is not None:
            cr = list(
                (self.w_stress*(
                        self.atoms.get_stress() - self.stress
                        )/self.stress).flatten()
                )
            r += cr
            w += list(self.w_stress*np.ones_like(self.stress).flatten())
        return r


### Specific structures

class FitDimer(Fit):

    __slots__ = [ 'D0', 'fmax', 'r0', 'w_D0', 'w_r0' ]

    def __init__(self, calc, par, els, D0, r0,
                 w_D0=1.0, w_r0=1.0,
                 vacuum=10.0, fmax=1e-6):
        Fit.__init__(self, calc, par)

        self.els = els

        self.D0 = D0
        self.r0 = r0

        self.w_D0 = sqrt(w_D0)/self.D0
        self.w_r0 = sqrt(w_r0)/self.r0

        self.calc = calc
        self.par = par

        self.fmax = fmax

        if type(self.els) == str:
            self.els = 2*[self.els]

        self.atoms = None

    def new_dimer(self):
        self.atoms = ase.Atoms(
            self.els,
            [[0.0, 0.0, 0.0],
             [0.0, 0.0, r0 ]],
            pbc = False)
        self.atoms.center(vacuum=vacuum)

    def set_calculator(self, calc):
        """
        Set the calculator, and relax the structure to its ground-state.
        """
        self.new_dimer()
        self.atoms.set_calculator(calc)
        ase.optimize.FIRE(self.atoms, logfile=_logfile).run(fmax=self.fmax,steps=10000)

    def get_distance(self):
        return self.atoms.get_distance(0, 1)

    def get_residuals(self, log=None):
        D0 = self.atoms.get_potential_energy()
        r0 = self.atoms.get_distance(0, 1)

        r_D0 = self.w_D0*(D0+self.D0)
        r_r0 = self.w_r0*(r0-self.r0)

        if log is not None:
            print('# %20s D0  = %20.10f eV    (%20.10f eV)    - %20.10f' \
                % ( 'Dimer', D0, -self.D0, r_D0 ), file=log)
            print('# %20s r0  = %20.10f A     (%20.10f A)     - %20.10f' \
                % ( '', r0, self.r0, r_r0 ), file=log)

        return r_D0, r_r0

class FitCubicCrystal(Fit):

    __slots__ = [ 'a0', 'calc', 'crystal', 'Ec', 'fmax', 'par', 'w_a0', 'w_Ec' ]

    def __init__(self, calc, par, els,
                 Ec, a0,
                 B=None, C11=None, C12=None, C44=None, Cp=None,SFE=None,
                 w_Ec=1.0, w_a0=1.0,
                 w_B=1.0, w_C11=1.0, w_C12=1.0, w_C44=1.0, w_Cp=1.0,w_SFE=1.0,
                 fmax=1e-6, eps=0.001,
                 ecoh_ref=None,
                 size=[1,1,1]):
        Fit.__init__(self, calc, par)

        self.els = els

        self.a0 = a0
        self.Ec = Ec
        
        self.SFE = SFE

        self.B = B
        self.C11 = C11
        self.C12 = C12
        self.C44 = C44
        self.Cp = Cp

        self.ecoh_ref = ecoh_ref

        self.w_a0 = sqrt(w_a0)/self.a0
        self.w_Ec = sqrt(w_Ec)/self.Ec

        if self.SFE is not None:
            self.w_SFE = sqrt(w_SFE)/self.SFE
        if self.B is not None:
            self.w_B = sqrt(w_B)/self.B
        if self.C11 is not None:
            self.w_C11 = sqrt(w_C11)/self.C11
        if self.C12 is not None:
            self.w_C12 = sqrt(w_C12)/self.C12
        if self.C44 is not None:
            self.w_C44 = sqrt(w_C44)/self.C44
        if self.Cp is not None:
            self.w_Cp = sqrt(w_Cp)/self.Cp

        self.size = size

        self.fmax = fmax
        self.eps = eps

        self.atoms = None

    def new_bulk(self):
        self.unitcell = self.crystal(
            self.els,
            latticeconstant  = self.a0,
            size             = [1, 1, 1]
            )
        self.atoms = self.unitcell.copy()
        self.atoms *= self.size
        self.atoms.translate([0.1, 0.1, 0.1])

    def set_calculator(self, calc):
        self.new_bulk()
        self.atoms.set_calculator(calc)
        ase.optimize.FIRE(
            ase.constraints.StrainFilter(self.atoms, mask=[1,1,1,0,0,0]),
            logfile=_logfile).run(fmax=self.fmax,steps=10000)
        a0 = self.get_lattice_constant()
        self.supercell = self.crystal(
            self.els,
            latticeconstant  = a0,
            size             = [3, 3, 6],
            directions=[[1,1,2],[-1,1,0],[-1,-1,1]],
            pbc=(1,1,0)
            )
        self.supercell.translate([0.1, 0.1, 0.1])
        self.supercell.set_calculator(calc)

    def get_lattice_constant(self):
        return np.sum(self.atoms.get_cell().diagonal())/np.sum(self.size)
      
    def get_SFE(self):
        E0 = self.supercell.get_potential_energy()/J*1e3
        S0 = self.supercell.get_cell()[0,0]*self.supercell.get_cell()[1,1]/m/m
        pos = self.supercell.get_positions()
        x,y=[],[]
        for i in range(10):
            pos1= np.copy(pos)
            pos1[:,0][pos[:,2]>self.supercell.get_cell()[2,2]/2-2]-=(self.supercell.get_cell()[0,0]/3)*0.08+(self.supercell.get_cell()[0,0]/3)*(1./50)*i
            self.supercell.set_positions(pos1)
            Es = self.supercell.get_potential_energy()/J*1e3
            x.append(0.08+(1./50)*i)
            y.append((Es/S0)-(E0/S0))
        GSF_fit = scipy.interpolate.InterpolatedUnivariateSpline(x,y)
        x_fit = np.linspace(0.08,0.08+(9./50),50)
        mins = argrelextrema(GSF_fit(x_fit),np.less)
        x_mins = x_fit[mins[0]]
        y_mins = GSF_fit(x_fit)[mins[0]]
        return y_mins[0]

    def get_C11(self):
        sxx0, syy0, szz0, syz0, szx0, sxy0  = self.atoms.get_stress()

        cell = self.atoms.get_cell()
        T = np.diag( [ self.eps, 0.0, 0.0 ] )
        self.atoms.set_cell( np.dot(np.eye(3)+T, cell), scale_atoms=True )
        sxx11, syy11, szz11, syz11, szx11, sxy11  = self.atoms.get_stress()
        self.atoms.set_cell(cell, scale_atoms=True)

        return (sxx11-sxx0)/self.eps

    def get_Cp(self):
        sxx0, syy0, szz0, syz0, szx0, sxy0  = self.atoms.get_stress()

        cell = self.atoms.get_cell()
        T = np.diag( [ self.eps, -self.eps, 0.0 ] )
        self.atoms.set_cell( np.dot(np.eye(3)+T, cell), scale_atoms=True )
        sxx12, syy12, szz12, syz12, szx12, sxy12  = self.atoms.get_stress()
        self.atoms.set_cell(cell, scale_atoms=True)

        return ((sxx12-sxx0)-(syy12-syy0))/(4*self.eps)

    def get_C44(self):
        sxx0, syy0, szz0, syz0, szx0, sxy0  = self.atoms.get_stress()

        cell = self.atoms.get_cell()
        T = np.array( [ [ 0.0, 0.5*self.eps, 0.5*self.eps ],
                        [ 0.5*self.eps, 0.0, 0.5*self.eps ],
                        [ 0.5*self.eps, 0.5*self.eps, 0.0 ] ] )
        self.atoms.set_cell( np.dot(np.eye(3)+T, cell), scale_atoms=True )
        sxx44, syy44, szz44, syz44, szx44, sxy44  = self.atoms.get_stress()
        self.atoms.set_cell(cell, scale_atoms=True)

        return (syz44+szx44+sxy44-syz0-szx0-sxy0)/(3*self.eps)
        
    def get_residuals(self, log=None):
        Ec = self.get_potential_energy()
        a0 = self.get_lattice_constant()

        if self.ecoh_ref is not None:
            syms = np.array(self.atoms.get_chemical_symbols())
            for el in set(syms):
                Ec -= (syms==el).sum()*self.ecoh_ref[el]

        Ec /= len(self.atoms)
        r_Ec = self.w_Ec*( Ec - self.Ec )
        r_a0 = self.w_a0*( a0 - self.a0 )

        if log is not None:
            print('# %20s Ec  = %20.10f eV/at.    (%20.10f eV/at.)    - %20.10f' \
                % ( '%s (%s)' % (self.unitcell.get_chemical_formula(),
                                 self.crystalstr), Ec, self.Ec, r_Ec ))
            print('# %20s a0  = %20.10f A     (%20.10f A)     - %20.10f' \
                % ( '', a0, self.a0, r_a0 ))

        r = [ r_Ec, r_a0 ]

        if self.SFE is not None:
            SFE = self.get_SFE()
        if self.B is not None or self.C11 is not None or self.C12 is not None:
            C11 = self.get_C11()
        if self.B is not None or self.Cp is not None or self.C12 is not None:
            Cp = self.get_Cp()
        if self.C44 is not None:
            C44 = self.get_C44()

        if self.SFE is not None:
            if SFE < 0:
                r_SFE = self.w_SFE*( SFE - self.SFE )*1000
            else:  
                r_SFE = self.w_SFE*( SFE - self.SFE )
            r += [ r_SFE ]
            if log is not None:
                print('# %20s SFE = %20.10f mJ/m**2   (%20.10f mJ/m**2)   - %20.10f' \
                    % ( '', SFE, self.SFE, r_SFE ))
        if self.B is not None:
            r_B = self.w_B*( (3*C11-4*Cp)/3 - self.B )
            r += [ r_B ]
            if log is not None:
                print('# %20s B   = %20.10f GPa   (%20.10f GPa)   - %20.10f' \
                    % ( '', (3*C11-4*Cp)/3/GPa, self.B/GPa, r_B ))
        if self.C11 is not None:
            r_C11 = self.w_C11*( C11 - self.C11 )
            r += [ r_C11 ]
            if log is not None:
                print('# %20s C11 = %20.10f GPa   (%20.10f GPa)   - %20.10f' \
                    % ( '', C11/GPa, self.C11/GPa, r_C11 ))
        if self.C12 is not None:
            r_C12 = self.w_C12*( C11-2*Cp - self.C12 )
            r += [ r_C12 ]
            if log is not None:
                print('# %20s C12 = %20.10f GPa   (%20.10f GPa)   - %20.10f' \
                    % ( '', (C11-2*Cp)/GPa, self.C12/GPa, r_C12 ))
        if self.C44 is not None:
            r_C44 = self.w_C44*( C44 - self.C44 )
            r += [ r_C44 ]
            if log is not None:
                print('# %20s C44 = %20.10f GPa   (%20.10f GPa)   - %20.10f' \
                    % ( '', C44/GPa, self.C44/GPa, r_C44 ))
        if self.Cp is not None:
            r_Cp = self.w_Cp*( Cp - self.Cp )
            r += [ r_Cp ]
            if log is not None:
                print('# %20s Cp  = %20.10f GPa   (%20.10f GPa)   - %20.10f' \
                    % ( '', Cp/GPa, self.Cp/GPa, r_Cp ))
        
        return r
      
class FitTetragonalCrystal(Fit):

    __slots__ = [ 'a0','c0', 'calc', 'crystal', 'Ec', 'fmax', 'par', 'w_a0','w_c0', 'w_Ec' ]

    def __init__(self, calc, par, els,
                 Ec, a0, c0, c_a=None,
                 B=None, C11=None, C12=None,C13=None, C33=None, C44=None, C66=None,SFE=None,
                 w_Ec=1.0, w_a0=1.0,w_c0=1.0, w_c_a=1.0,
                 w_B=1.0, w_C11=1.0, w_C12=1.0,w_C13=1.0,w_C33=1.0,w_C44=1.0,w_C66=1.0, w_Cp=1.0,w_SFE=None,
                 fmax=1e-6, eps=0.001,
                 ecoh_ref=None,
                 size=[1,1,1]):
        Fit.__init__(self, calc, par)

        self.els = els

        self.a0 = a0
        self.c0 = c0
        self.Ec = Ec
        self.c_a = c_a

        self.SFE = SFE
        
        self.B = B
        self.C11 = C11
        self.C12 = C12
        self.C13 = C13
        self.C44 = C44
        self.C33 = C33
        self.C66 = C66


        self.ecoh_ref = ecoh_ref

        self.w_a0 = sqrt(w_a0)/self.a0
        self.w_c0 = sqrt(w_c0)/self.c0
        
        if self.Ec is not None:
            self.w_Ec = sqrt(w_Ec)/self.Ec
        if self.c_a is not None:
            self.w_c_a = sqrt(w_c_a)/self.c_a
        if self.SFE is not None:
            self.w_SFE = sqrt(w_SFE)/self.SFE
        if self.B is not None:
            self.w_B = sqrt(w_B)/self.B
        if self.C11 is not None:
            self.w_C11 = sqrt(w_C11)/self.C11
        if self.C12 is not None:
            self.w_C12 = sqrt(w_C12)/self.C12
        if self.C13 is not None:
            self.w_C13 = sqrt(w_C13)/self.C13
        if self.C33 is not None:
            self.w_C33 = sqrt(w_C33)/self.C33
        if self.C44 is not None:
            self.w_C44 = sqrt(w_C44)/self.C44
        if self.C66 is not None:
            self.w_C66 = sqrt(w_C66)/self.C66

        self.size = size

        self.fmax = fmax
        self.eps = eps

        self.atoms = None
    
    def new_bulk(self):
        self.unitcell = self.crystal(
            self.els,
            latticeconstant  = [self.a0,self.c0],
            size             = [1, 1, 1]
            )
        self.atoms = self.unitcell.copy()
        self.atoms *= self.size
        self.atoms.translate([0.1, 0.1, 0.1])

    def set_calculator(self, calc):
        self.new_bulk()
        self.atoms.set_calculator(calc)
        ase.optimize.FIRE(
            ase.constraints.StrainFilter(self.atoms, mask=[1,1,1,1,1,1]),
            logfile=_logfile).run(fmax=self.fmax,steps=10000)
        a0,c0 = self.get_lattice_constant()
        self.supercell = self.crystal(
            self.els,
            latticeconstant  = [a0,c0],
            size             = [3, 3, 6],
            directions=[[1,1,2],[-1,1,0],[-1,-1,1]],
            pbc=(1,1,0)
            )
        self.supercell.translate([0.1, 0.1, 0.1])
        self.supercell.set_calculator(calc)

    def get_SFE(self):
        E0 = self.supercell.get_potential_energy()/J*1e3
        S0 = self.supercell.get_cell()[0,0]*self.supercell.get_cell()[1,1]/m/m
        pos = self.supercell.get_positions()
        x,y=[],[]
        for i in range(10):
            pos1= np.copy(pos)
            pos1[:,0][pos[:,2]>self.supercell.get_cell()[2,2]/2-2]-=(self.supercell.get_cell()[0,0]/3)*0.05+(self.supercell.get_cell()[0,0]/3)*(1./50)*i
            self.supercell.set_positions(pos1)
            Es = self.supercell.get_potential_energy()/J*1e3
            x.append(0.05+(1./50)*i)
            y.append((Es/S0)-(E0/S0))
        GSF_fit = scipy.interpolate.InterpolatedUnivariateSpline(x,y)
        x_fit = np.linspace(0.05,0.05+(9./50),50)
        mins = argrelextrema(GSF_fit(x_fit),np.less)
        x_mins = x_fit[mins[0]]
        y_mins = GSF_fit(x_fit)[mins[0]]
        return y_mins[0]
      
    def get_lattice_constant(self):
        return np.sum(self.atoms.get_cell().diagonal()[:2])/np.sum(self.size[:2]),self.atoms.get_cell().diagonal()[2]/self.size[2]

    def get_C33(self):
        e0 = self.atoms.get_potential_energy()
        v0 = self.atoms.get_volume()
        
        cell = self.atoms.get_cell()
        T = np.array( [ [ 0.0, 0.0, 0.0 ],
                        [ 0.0, 0.0, 0.0 ],
                        [ 0.0, 0.0, self.eps ] ] )
        self.atoms.set_cell( np.dot(np.eye(3)+T, cell), scale_atoms=True )
        e = self.atoms.get_potential_energy()
        self.atoms.set_cell(cell, scale_atoms=True)
        
        
        return (2*(e-e0))/(v0*self.eps**2)
      
    def get_C44(self):
        e0 = self.atoms.get_potential_energy()
        v0 = self.atoms.get_volume()
        
        cell = self.atoms.get_cell()
        T = np.array( [ [ 0.0, 0.0, self.eps ],
                        [ 0.0, 0.0, self.eps ],
                        [ self.eps, self.eps, self.eps**2 ] ] )
        self.atoms.set_cell( np.dot(np.eye(3)+T, cell), scale_atoms=True )
        e = self.atoms.get_potential_energy()
        self.atoms.set_cell(cell, scale_atoms=True)
        
        
        return (e-e0)/(4*v0*self.eps**2)
      
    def get_C66(self):
        e0 = self.atoms.get_potential_energy()
        v0 = self.atoms.get_volume()
        
        cell = self.atoms.get_cell()
        T = np.array( [ [ (1.0+self.eps**2)**0.5, 0.0, 0.0 ],
                        [ 0.0, (1.0-self.eps**2)**0.5, 0.0 ],
                        [ 0.0, 0.0, 1.0 ] ] )
        self.atoms.set_cell( np.dot(np.eye(3)+T, cell), scale_atoms=True )
        e = self.atoms.get_potential_energy()
        self.atoms.set_cell(cell, scale_atoms=True)
        
        
        return (e-e0)/(2*v0*self.eps**2)
      
    def get_D1(self):
        e0 = self.atoms.get_potential_energy()
        v0 = self.atoms.get_volume()
        
        cell = self.atoms.get_cell()
        T = np.array( [ [ self.eps, 0.0, 0.0 ],
                        [ 0.0, self.eps, 0.0 ],
                        [ 0.0, 0.0, 0.0 ] ] )
        self.atoms.set_cell( np.dot(np.eye(3)+T, cell), scale_atoms=True )
        e = self.atoms.get_potential_energy()
        self.atoms.set_cell(cell, scale_atoms=True)
        
        
        return (e-e0)/(v0*self.eps**2)

    def get_D2(self):
        e0 = self.atoms.get_potential_energy()
        v0 = self.atoms.get_volume()
        
        cell = self.atoms.get_cell()
        T = np.array( [ [ 1.0+self.eps, 0.0, 0.0 ],
                        [ 0.0, 1.0+self.eps, 0.0 ],
                        [ 0.0, 0.0, 1.0/(1+self.eps)**2 ] ] )
        self.atoms.set_cell( np.dot(T, cell), scale_atoms=True )
        e = self.atoms.get_potential_energy()
        self.atoms.set_cell(cell, scale_atoms=True)
        
        
        return (e-e0)/(v0*self.eps**2)
      
    def get_D4(self):
        e0 = self.atoms.get_potential_energy()
        v0 = self.atoms.get_volume()
        
        cell = self.atoms.get_cell()
        T = np.array( [ [ ((1.0+self.eps)/(1.0-self.eps))**0.5, 0.0, 0.0 ],
                        [ 0.0, ((1.0-self.eps)/(1.0+self.eps))**0.5, 0.0 ],
                        [ 0.0, 0.0, 1.0 ] ] )
        self.atoms.set_cell( np.dot(T, cell), scale_atoms=True )
        e = self.atoms.get_potential_energy()
        self.atoms.set_cell(cell, scale_atoms=True)
        
        
        return (e-e0)/(v0*self.eps**2)
      
        
    def get_residuals(self, log=None):
        Ec = self.get_potential_energy()
        a0,c0 = self.get_lattice_constant()

        if self.ecoh_ref is not None:
            syms = np.array(self.atoms.get_chemical_symbols())
            for el in set(syms):
                Ec -= (syms==el).sum()*self.ecoh_ref[el]

        Ec /= len(self.atoms)

        r_a0 = self.w_a0*( a0 - self.a0 )
        r_c0 = self.w_c0*( c0 - self.c0 )

        r_Ec = self.w_Ec*( Ec - self.Ec )
        
        if log is not None:
            print('# %20s Ec  = %20.10f eV/at.    (%20.10f eV/at.)    - %20.10f' \
                % ( '%s (%s)' % (self.unitcell.get_chemical_formula(),
                                 self.crystalstr), Ec, self.Ec, r_Ec ))
            print('# %20s a0  = %20.10f A     (%20.10f A)     - %20.10f' \
                % ( '', a0, self.a0, r_a0 ))
            print('# %20s c0  = %20.10f A     (%20.10f A)     - %20.10f' \
                % ( '', c0, self.c0, r_c0 ))

        r = [ r_a0 ,r_c0, r_Ec]
        
        if self.SFE is not None:
            SFE = self.get_SFE()
        if self.B is not None or self.C11 is not None or self.C12 is not None or self.C13 is not None or self.C33 is not None:
            Czz = self.get_D2()
        if self.B is not None or self.C11 is not None or self.C12 is not None:
            Cp = self.get_D1()
        if self.B is not None or self.C11 is not None or self.C12 is not None:
            Cm = self.get_D4()
        if self.C33 is not None:
            C33 = self.get_C66()
        if self.C44 is not None:
            C44 = self.get_C66()
        if self.C66 is not None:
            C66 = self.get_C66()


        if self.SFE is not None:
            if SFE < 0:
                r_SFE = self.w_SFE*( SFE - self.SFE )*1000
            else:  
                r_SFE = self.w_SFE*( SFE - self.SFE )
            r += [ r_SFE ]
            if log is not None:
                print('# %20s SFE = %20.10f mJ/m**2   (%20.10f mJ/m**2)   - %20.10f' \
                    % ( '', SFE, self.SFE, r_SFE ))
        if self.c_a is not None:
            if (self.c_a < 1 and c0/a0 < 1) or (self.c_a > 1 and c0/a0 > 1) :
                r_c_a = self.w_c_a*( c0/a0 - self.c_a )
            else:
                r_c_a = self.w_c_a*( c0/a0 - self.c_a )*1000
            r += [ r_c_a ]
            if log is not None:
                print('# %20s c/a   = %20.10f   (%20.10f )   - %20.10f' \
                    % ( '', c0/a0 , self.c_a, r_c_a ))
        if self.B is not None:
            r_B = self.w_B*( (C33*Cp-2*C13**2)/Czz - self.B )
            r += [ r_B ]
            if log is not None:
                print('# %20s B   = %20.10f GPa   (%20.10f GPa)   - %20.10f' \
                    % ( '', (C33*Cp-2*C13**2)/Czz/GPa, self.B/GPa, r_B ))
        if self.C11 is not None:
            r_C11 = self.w_C11*( (Cp+Cm)/2 - self.C11 )
            r += [ r_C11 ]
            if log is not None:
                print('# %20s C11 = %20.10f GPa   (%20.10f GPa)   - %20.10f' \
                    % ( '', (Cp+Cm)/2/GPa, self.C11/GPa, r_C11 ))
        if self.C12 is not None:
            r_C12 = self.w_C12*( (Cp-Cm)/2 - self.C12 )
            r += [ r_C12 ]
            if log is not None:
                print('# %20s C12 = %20.10f GPa   (%20.10f GPa)   - %20.10f' \
                    % ( '', (Cp-Cm)/2/GPa, self.C12/GPa, r_C12 ))
        if self.C13 is not None:
            r_C13 = self.w_C13*( -(Czz-Cp-2*C33)/4 - self.C13 )
            r += [ r_C13 ]
            if log is not None:
                print('# %20s C13 = %20.10f GPa   (%20.10f GPa)   - %20.10f' \
                    % ( '', -(Czz-Cp-2*C33)/4/GPa, self.C13/GPa, r_C13 ))
        if self.C33 is not None:
            r_C33 = self.w_C33*( C33 - self.C33 )
            r += [ r_C33 ]
            if log is not None:
                print('# %20s C33 = %20.10f GPa   (%20.10f GPa)   - %20.10f' \
                    % ( '', C33/GPa, self.C33/GPa, r_C33 ))
        if self.C44 is not None:
            r_C44 = self.w_C44*( C44 - self.C44 )
            r += [ r_C44 ]
            if log is not None:
                print('# %20s C44 = %20.10f GPa   (%20.10f GPa)   - %20.10f' \
                    % ( '', C44/GPa, self.C44/GPa, r_C44 ))
        if self.C66 is not None:
            r_C66 = self.w_C66*( C66 - self.C66 )
            r += [ r_C66 ]
            if log is not None:
                print('# %20s C66 = %20.10f GPa   (%20.10f GPa)   - %20.10f' \
                    % ( '', C66/GPa, self.C66/GPa, r_C66 ))

        
        return r

class FitHexagonalCrystal(Fit):

    __slots__ = [ 'a0', 'c0', 'calc', 'crystal', 'Ec', 'fmax', 'par', 'w_a0',
                  'w_Ec' ]

    def __init__(self, calc, par, els,
                 Ec, a0, c0,
                 w_Ec = 1.0, w_a0 = 1.0,
                 fmax = 0.01):
        Fit.__init__(self, calc, par)

        self.els = els

        self.Ec = Ec
        self.a0 = a0
        self.c0 = c0

        self.w_Ec = sqrt(w_Ec)/self.Ec
        self.w_a0 = sqrt(w_a0)/self.a0

        self.fmax = fmax

        self.atoms = None

    def new_bulk(self):
        self.atoms = self.crystal(
            self.els,
            latticeconstant  = [self.a0, self.c0],
            size             = [1, 1, 1]
            )
        self.atoms.translate([0.1, 0.1, 0.1])

    def set_calculator(self, calc):
        self.new_bulk()
        self.atoms.set_calculator(calc)
        ase.optimize.FIRE(
            ase.constraints.StrainFilter(self.atoms, mask=[1,1,0,0,0,0]),
            logfile=_logfile).run(fmax=self.fmax,steps=10000)

    def get_lattice_constant(self):
        cx, cy, cz = self.atoms.get_cell()
        
        return ( sqrt(np.dot(cx, cx)) + sqrt(np.dot(cy, cy)) )/2

    def get_residuals(self, log=None):
        Ec = self.get_potential_energy()/len(self.atoms)
        a0 = self.get_lattice_constant()

        r_Ec = self.w_Ec*( Ec + self.Ec )
        r_a0 = self.w_a0*( a0 - self.a0 )

        if log is not None:
            print('# %20s Ec  = %20.10f eV/at.    (%20.10f eV/at.)    - %20.10f' \
                % ( 'Crystal (%s)' % self.crystalstr, Ec, -self.Ec, r_Ec ))
            print('# %20s a0  = %20.10f A     (%20.10f A)     - %20.10f' \
                % ( '', a0, self.a0, r_a0 ))

        r = [ r_Ec, r_a0 ]

        return r

class FitSurface(Fit):

    __slots__ = [ 'a0', 'calc', 'crystal', 'Ec', 'fmax', 'par', 'w_a0', 'w_Ec' ]

    def __init__(self, calc, par, els, crystal,
                 Esurf,
                 w_Esurf = 1.0):
        self.Esurf = Esurf

        self.w_Esurf = sqrt(w_Esurf)

        self.els = els

        self.calc = calc
        self.par = par

        self.crystal = crystal

    def set_calculator(self, calc):
        self.atoms, self.ncells = \
            self.new_surface(self.crystal.get_lattice_constant())
        ase.io.write('%s.cfg' % self.surfstr, self.atoms)
        self.atoms.set_calculator(calc)

    def get_surface_energy(self):
        return ( self.atoms.get_potential_energy() -
                 self.crystal.get_cohesive_energy()*len(self.atoms) ) \
                 /(2*self.ncells)

    def get_residuals(self, log=None):
        Esurf = self.get_surface_energy()
        sx, sy, sz = self.atoms.get_cell().diagonal()
        tar_Esurf = self.Esurf*(sx*sy)/self.ncells
        r_Esurf = self.w_Esurf*( Esurf - tar_Esurf )/tar_Esurf

        if log is not None:
            print('# %20s Es  = %20.10f eV    (%20.10f eV)    - %20.10f' \
                % ( 'Surface(%s)' % self.surfstr, Esurf,
                    tar_Esurf, r_Esurf ))
            print('# %20s       %20.10f J/m^2 (%20.10f J/m^2)' \
                % ( '',
                    Esurf*self.ncells*Jm2/(sx*sy),
                    self.Esurf*Jm2 ))
        
        return [ r_Esurf ]

###

class FitSC(FitCubicCrystal):
    crystal = cubic.SimpleCubic
    crystalstr = 'sc'

class FitBCC(FitCubicCrystal):
    crystal = cubic.BodyCenteredCubic
    crystalstr = 'bcc'

class FitFCC(FitCubicCrystal):
    crystal = cubic.FaceCenteredCubic
    crystalstr = 'fcc'

class FitB2(FitCubicCrystal):
    crystal = compounds.B2
    crystalstr = 'B2'

class FitL1_0(FitTetragonalCrystal):
    crystal = compounds.L1_0
    crystalstr = 'L1_0'

class FitL1_2(FitCubicCrystal):
    crystal = compounds.L1_2
    crystalstr = 'L1_2'

class FitDiamond(FitCubicCrystal):
    crystal = cubic.Diamond
    crystalstr = 'dia'

class FitGraphite(FitHexagonalCrystal):
    crystal = hexagonal.Graphite
    crystalstr = 'gra'

class FitGraphene(FitHexagonalCrystal):
    crystal = hexagonal.Graphene
    crystalstr = 'grp'

class FitDiamond100(FitSurface):
    surfstr = 'dia-100'

    def new_surface(self, a0):
        a = cubic.Diamond(
            self.els,
            latticeconstant = a0,
            directions = [ [ 1,0,0 ],
                           [ 0,1,0 ],
                           [ 0,0,1 ] ],
            size = [ 1, 1, 5 ])
        a.translate([0.1,0.1,0.1])
        sx, sy, sz = a.get_cell().diagonal()
        a.set_cell([sx,sy,2*sz])

        return a, 2

class FitDiamond110(FitSurface):
    surfstr = 'dia-110'

    def new_surface(self, a0):
        a = cubic.Diamond(
            self.els,
            latticeconstant = a0,
            directions = [ [ 1, 0,0 ],
                           [ 0, 1,1 ],
                           [ 0,-1,1 ] ],
            size = [ 1, 1, 5 ])
        a.translate([0.1,0.1,0.1])
        sx, sy, sz = a.get_cell().diagonal()
        a.set_cell([sx,sy,2*sz])

        return a, 1

class FitDiamond111(FitSurface):
    surfstr = 'dia-111'
 
    def new_surface(self, a0):
        a = cubic.Diamond(
            self.els,
            latticeconstant = a0,
            directions = [ [ 2,-1,-1 ],
                           [ 0, 1,-1 ],
                           [ 1, 1, 1 ] ],
            size = [ 1, 1, 5 ])
        a.translate([0.1,0.1,0.1+a0/4])
        a.set_scaled_positions(a.get_scaled_positions())
        sx, sy, sz = a.get_cell().diagonal()
        a.set_cell([sx,sy,2*sz])

        return a, 2



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

"""
Compute deformation gradient tensor and D^2_min measure for non-affine
displacements.
See: Falk, Langer, Phys. Rev. E 57, 7192 (1998)
"""

import numpy as np

from matscipy.neighbours import mic, neighbour_list

###

def get_XIJ(nat, i_now, dr_now, dr_old):
    """
    Calculates the X_{ij} matrix
    """
    # Do an element-wise outer product
    dr_dr = dr_now.reshape(-1,3,1)*dr_old.reshape(-1,1,3)

    xij = np.zeros([nat,3,3])
    for i in range(3):
        for j in range(3):
            # For each atom, sum over all neighbors
            xij[:,i,j] = np.bincount(i_now, weights=dr_dr[:,i,j])

    return xij


def get_YIJ(nat, i_now, dr_old):
    """
    Calculates the Y_{ij} matrix
    """
    # Just do an element-wise outer product
    dr_dr = dr_old.reshape(-1,3,1)*dr_old.reshape(-1,1,3)

    yij = np.zeros([nat,3,3])
    for i in range(3):
        for j in range(3):
            # For each atom, sum over all neighbors
            yij[:,i,j] = np.bincount(i_now, weights=dr_dr[:,i,j])

    return yij


def array_inverse(A):
    """
    Compute inverse for each matrix in a list of matrices.
    This is faster than calling numpy.linalg.inv for each matrix.
    """
    A = np.ascontiguousarray(A, dtype=float)
    b = np.identity(A.shape[2], dtype=A.dtype)

    n_eq = A.shape[1]
    n_rhs = A.shape[2]
    pivots = np.zeros(n_eq, np.intc)
    identity = np.eye(n_eq)
    def lapack_inverse(a):
        b = np.copy(identity)
        pivots = np.zeros(n_eq, np.intc)
        results = np.linalg.lapack_lite.dgesv(n_eq, n_rhs, a, n_eq, pivots, b,
                                              n_eq, 0)
        if results['info'] > 0:
            raise np.linalg.LinAlgError('Singular matrix')
        return b

    return np.array([lapack_inverse(a) for a in A])


def get_delta_plus_epsilon_dgesv(nat, i_now, dr_now, dr_old):
    """
    Calculate delta_ij+epsilon_ij, i.e. the deformation gradient matrix
    """
    XIJ = get_XIJ(nat, i_now, dr_now, dr_old)
    YIJ = get_YIJ(nat, i_now, dr_old)

    YIJ_invert = array_inverse(YIJ)

    # Perform sum_k X_ik Y_jk^-1
    epsilon = np.sum(XIJ.reshape(-1,3,1,3)*YIJ_invert.reshape(-1,1,3,3), axis=3)

    return epsilon


def get_delta_plus_epsilon(nat, i_now, dr_now, dr_old):
    """
    Calculate delta_ij+epsilon_ij, i.e. the deformation gradient matrix
    """
    epsilon = []
    for i in range(nat):
        mask = i_now==i
        x, residuals, rank, s = np.linalg.lstsq(dr_old[mask], dr_now[mask])
        epsilon += [ x.T ]

    return np.array(epsilon)


def get_D_square_min(atoms_now, atoms_old, i_now, j_now, delta_plus_epsilon=None):
    """
    Calculate the D^2_min norm of Falk and Langer
    """
    nat = len(atoms_now)
    assert len(atoms_now) == len(atoms_old)

    pos_now = atoms_now.positions
    pos_old = atoms_old.positions

    # Compute current and old distance vectors. Note that current distance
    # vectors cannot be taken from the neighbor calculation, because neighbors
    # are calculated from the sheared cell while these distance need to come
    # from the unsheared cell. Taking the distance from the unsheared cell
    # make periodic boundary conditions (and flipping of cell) a lot easier.
    dr_now = mic(pos_now[i_now] - pos_now[j_now], atoms_now.cell)
    dr_old = mic(pos_old[i_now] - pos_old[j_now], atoms_old.cell)

    # Sanity check: Shape needs to be identical!
    assert dr_now.shape == dr_old.shape

    if delta_plus_epsilon is None:
        # Get minimum strain tensor
        delta_plus_epsilon = get_delta_plus_epsilon(nat, i_now, dr_now, dr_old)

    # Spread epsilon out for each neighbor index
    delta_plus_epsilon_n = delta_plus_epsilon[i_now]

    # Compute D^2_min (residual of the least squares fit)
    residual_n = np.sum(
        (
        dr_now-
        np.sum(delta_plus_epsilon_n.reshape(-1,3,3)*dr_old.reshape(-1,1,3),
               axis=2)
        )**2,
        axis=1)

    # For each atom, sum over all neighbors
    residual = np.bincount(i_now, weights=residual_n)

    return delta_plus_epsilon, residual


def atomic_strain(atoms_now, atoms_old, cutoff=None, neighbours=None):
    """
    Calculate deformation gradient tensor and D^2_min measure for non-affine
    displacements.
    See: Falk, Langer, Phys. Rev. B 57, 7192 (1998)

    Parameters:
    -----------
    atoms_now : ase.Atoms
        Current atomic configuration
    atoms_old : ase.Atoms
        Reference atomic configuration
    cutoff : float
        Neighbor list cutoff.
    neighbours : ( array_like, array_like )
        Neighbor list. Automatically computed if not provided.

    Returns:
    --------
    delta_plus_epsilon : array
        3x3 deformation gradient tensor for each atom.
    residual : array
        D^2_min norm for each atom
    """

    if neighbours is None:
        if cutoff is None:
            raise ValueError('Please provide either neighbor list or neighbor '
                             'list cutoff.')

        # Get neighbours
        i_now, j_now = neighbour_list("ij", atoms_now, cutoff)
    elif cutoff is not None:
        raise ValueError('Please provide either neighbor list or neighbor '
                         'list cutoff, not both.')
    else:
        i_now, j_now = neighbours

    # Get strain gradient tensor and D square values
    delta_plus_epsilon, residual = get_D_square_min(atoms_now, atoms_old, i_now,
                                                    j_now)

    return delta_plus_epsilon, residual



from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np

from .qm_cluster_tools.qm_flagging_tool import QMFlaggingTool
from .qm_cluster_tools.qm_clustering_tool import QMClusteringTool


class QMCluster(object):
    """This is a class responsible for managing the QM clusters in the simulation.

    It acts as a mediator between
    -----------------------------
        QM flagginig module
        QM clustering module
        neighbour list
    and contains interface to those classes

    Attributes
    ----------
    clustering_module : matscipy.calculators.mcfm.qmClusterModule.QMClusteringTool
        module responsible for carving a qm cluster
    flagging_module : matscipy.calculators.mcfm.qmClusterModule.QMFlaggingTool
        module responsible for flagging atoms
    neighbour_list : matscipy.calculators.mcfm.neighbour_list_mcfm.NeighborListBase
        object holding the neighbour list
    verbose : int
        Set verbosity level
    """

    def __init__(self, special_atoms_list=[], verbose=0):
        """This is a class responsible for managing the QM clusters in the simulation.

        It acts as a mediator between
        -----------------------------
            QM flagginig module
            QM clustering module
            neighbour list
        and contains interface to those classes

        Parameters
        ----------
        special_atoms_list : list of ints (atomic indices)
            In case a group of special atoms are specified (special molecule),
            If one of these atoms is in the buffer region, the rest are also added to it.
        verbose : int
            verbosity level to be passed to other objects
        """
        self.flagging_module = None
        self.clustering_module = None
        self.neighbour_list = None

        self.special_atoms_list = special_atoms_list
        self.verbose = verbose

    def attach_neighbour_list(self, neighbour_list):
        """attach a neighbour list"""
        self.neighbour_list = neighbour_list

    def attach_flagging_module(self, **kwargs):
        """Initialize and attach matscipy.calculators.mcfm.QMFlaggingTool
        The function calls the class initializer with given parameters"""
        self.flagging_module = QMFlaggingTool(mediator=self, **kwargs)

    def attach_clustering_module(self, **kwargs):
        """Initialize and attach matscipy.calculators.mcfm.QMClusteringTool
        The function calls the class initializer with given parameters"""
        self.clustering_module = QMClusteringTool(mediator=self, **kwargs)

    def reset_energized_list(self):
        """Reset old_energized_atoms list in flaggingModule to facilitate
        MCFM potential warmup"""
        self.flagging_module.old_energized_list = []

    def update_qm_region(self, *args, **kwargs):
        """Interface to
        self.flagging_module.update_qm_region(self,
                                             atoms,
                                             potential_energies=None,
                                             )"""
        return self.flagging_module.update_qm_region(*args, **kwargs)

    def carve_cluster(self, *args, **kwargs):
        """Interface to
        self.clustering_module.carve_cluster(self,
                                            atoms,
                                            core_qm_list,
                                            buffer_hops=10)"""
        return self.clustering_module.carve_cluster(*args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np

from ase.calculators.calculator import Calculator
from .mcfm_parallel import mcfm_parallel_control as mcfm_parallel_control
from .cluster_data import ClusterData
import ase.io


class MultiClusterForceMixingPotential(Calculator):
    """Subclass of ASE.Calculator facilitating a support for multiple
    QM clusters. It utilizes the given classical_calculator and qm_calculator to initialize
    an instace of ForceMixingPotential

    Extends:
        Calculator

    Variables:
        implemented_properties {list} -- ["energy", "forces", "potential_energies", "stress"]
        all_changes {list} -- ['positions', 'numbers', 'cell', 'pbc']
    """

    implemented_properties = ["energy", "forces", "potential_energies", "stress"]
    all_changes = ['positions', 'numbers', 'cell', 'pbc']

    def __init__(self, atoms=None, classical_calculator=None, qm_calculator=None,
                 qm_cluster=None, forced_qm_list=None, change_bonds=True,
                 calculate_errors=False, calculation_always_required=False,
                 buffer_hops=10, verbose=0, enable_check_state=True):
        """Initialize a generic ASE potential without any calculating power,
        This is only to have access to the necessary functions, all the
        evaluations will be performes in self.mm_pot and self.qm_calculator

        Parameters
        ----------
        atoms : ASE.atoms
            atoms object
        classical_calculator : AE.calculator
            classical calculator
        qm_calculator : ASE.calculator
            qm calculator
        qm_cluster : matscipy.calculators.mcfm.qm_cluster
            flagging/cluster carving utility
        forced_qm_list : list
            add this list to enforce a set of atoms for qm treatment
        change_bonds : bool
            call the classical potential to update topology
        calculate_errors : bool
            evaluate errors after each step
        calculation_always_required : bool
            as name
        buffer_hops : int
            number of neighbours hop used to construct the core QM region
        verbose : int
            For now verbose levels are:
            0 - nothing is printed
            1 - More data is added to Atoms object
            10 - Calculate steps are listed
            100 - Information about specific QM clusters
            (the default is 0)
        enable_check_state : bool
            Save the atoms after each evaluation to enable meth::check_state
        """

        # Set the verbose status
        self.verbose = verbose
        self.debug_cluster_carving = False

        # Set storing atoms - slows down evaluation but enables check_state funtion
        self.enable_check_state = enable_check_state

        # Flag for warmup
        self.warmup = False

        # Init ASE calculator as a parent class
        self._calc_args = {}
        self._default_properties = []
        self.calculation_always_required = calculation_always_required
        Calculator.__init__(self)

        # If an atoms objct has been specified, attach a copy to the calculator to facilitate
        # the proper use of meth:check_state()
        if atoms is not None:
            self.atoms = atoms.copy()
            atoms.set_calculator(self)

        # Set some flags and values
        self.errors = {}
        self.calculate_errors = calculate_errors
        self.change_bonds = change_bonds
        self.buffer_hops = buffer_hops
        self.conserve_momentum = False
        self.long_range_weight = 0.0
        self.doParallel = True

        # Flag for QM debugging
        self.debug_qm_calculator = False

        # Set the cluster carving object
        self.qm_cluster = qm_cluster
        if forced_qm_list is None:
            self.forced_qm_list = None
        else:
            self.forced_qm_list = [forced_qm_list]
        self.cluster_list = []
        self.cluster_data_list = None

        # Set qm and mm calculators
        self.classical_calculator = classical_calculator
        self.qm_calculator = qm_calculator

        # Set up writing clusters
        self.clusterDebug_cluster_atoms = None
        if (self.verbose >= 100):
            self.debug_cluster_carving = True
            self.clusterDebug_cluster_atoms = open("clusterDebug_cluster_atoms.xyz", "w")
        self.clusterDebug_full_structure = None
        if (self.verbose >= 100):
            self.debug_cluster_carving = True
            self.clusterDebug_full_structure = open("clusterDebug_full_structure.xyz", "w")

    def calculate(self, atoms=None, properties=['energy'],
                  system_changes=all_changes):
        """Calculate selected properties of the given Atoms object.
        Initially, a classical potential is called to evaluate potential
        energies for each atom and afterwards a qm_cluster object is employed
        to analyze them. If no atom is flagged for QM treatment, classical forces
        are returned. In case some atoms are flagged for QM treatment
        each Qm cluster is independently send to a qmmm potential to evaluate
        more accurate forces. The results of qmmm evaluations are used to modify
        the classical forces and the final array is given

        results are
        -----------
        energy = potential energy from classical evaluation
        potential energies = pot energy per atom from classical evaluation
        forces = classical or qmmm forces depending on whether any atoms are flagged

        Parameters
        ----------
        atoms : ASE.atoms
            atoms object
        properties : list
            properties ot evaluate
        system_changes : TYPE
            changes in the system

        Raises
        ------
        AttributeError
            Must provide an atoms object
        """

        Calculator.calculate(self, atoms=atoms, properties=properties, system_changes=system_changes)
        # Check for atoms and if present wrap them to their cell
        if atoms is None:
            raise AttributeError("No atoms object provided")
        else:
            pass

        self.print_message("\nCalculation with MCFM potential!", limit=10)
        self.print_message("Calculating parameters using classical potential", limit=10)
        forces, potential_energy, potential_energies =\
            self.produce_classical_results(atoms=atoms)

        self.results["classical_forces"] = forces.copy()

        self.print_message(
            "Update the qm_cluster based on potential energies obtained from mm_pot calculation", limit=10)

        # If in warmup mode, do not use the clusters
        self.produce_qm_clusters(atoms,
                                 potential_energies=potential_energies)

        if (self.warmup):
            self.print_message("Warmup mode: not using any clusters", limit=10)
            self.cluster_list = []
            self.qm_cluster.reset_energized_list()
        self.print_message("Cluster list", limit=100)
        self.print_message(self.cluster_list, limit=100)

        # Perform parallel QM calculations on each cluster
        cluster_data_list = np.empty(len(self.cluster_list), dtype=object)
        if (len(cluster_data_list) > 0):
            if (self.doParallel):
                mcfm_parallel_control.get_cluster_data(atoms=atoms,
                                                       clusterData=cluster_data_list,
                                                       mcfm_pot=self)
            else:
                for i, cluster in enumerate(self.cluster_list):
                    # Evaluate qm cluster
                    cluster_data_list[i] = self.evaluate_qm_cluster_serial(
                        atoms=atoms, cluster=cluster, clusterNumber=i)

        # ------ Attach the cluster data list
        self.cluster_data_list = cluster_data_list

        # Stitch the forces using data from cluster_data_list
        forces = self.combine_qm_mm_forces(atoms=atoms,
                                           forces=forces,
                                           cluster_data_list=cluster_data_list)

        # Create the full QM list and QM mask
        full_qm_atoms_list = [item for sublist in self.cluster_list for item in sublist]
        full_qm_atoms_mask = np.zeros(len(atoms), dtype=bool)
        full_qm_atoms_mask[full_qm_atoms_list] = True

        # If the potential can update topology, do it.
        do_bond_change = (not (len(full_qm_atoms_list) == 0)) and\
            (self.change_bonds is True) and\
            hasattr(self.classical_calculator, "update_topology")
        if do_bond_change:
            self.print_message("Updating Topology!", limit=10)
            self.classical_calculator.update_topology(full_qm_atoms_list)

        # Mark atoms that are treated quantm mmechanically for more comprehensive ooutput
        self.attach_hybrid_data(
            atoms=atoms, full_qm_atoms_mask=full_qm_atoms_mask, cluster_data=cluster_data_list)

        # Compute stress
        if "stress" in properties:
            self.results["stress"] = self.compute_stress(atoms, forces)

        # Attach the updated version of atoms so that check-state would work properly.
        if self.enable_check_state:
            self.atoms = atoms.copy()

        self.results["forces"] = forces
        self.results["potential_energy"] = potential_energy
        self.results["energy"] = potential_energy
        self.results["potential_energies"] = potential_energies

        if (self.calculate_errors):
            self.evaluate_errors(atoms=atoms)

    def produce_classical_results(self, atoms=None):
        """Call the classical potential ot obtain forces, potential energy
        and potential energies per atom

        Parameters
        ----------
        atoms : ASE.atoms
            atoms object

        Returns
        -------
        forces : np.array
            Atomic forces
        potential_energy : np.array
            Potential energy of the system
        potential_energies : np.array
            Per atom potential energies
        """
        # Get forces
        forces = self.classical_calculator.get_forces(atoms)

        # Get potential energies
        if "potential_energies" in self.classical_calculator.results:
            potential_energies = self.classical_calculator.results["potential_energies"]
        else:
            potential_energies = self.classical_calculator.get_potential_energies(atoms)

        # Get total potential energy
        # (summing over individual contributions isusually faster then a calculation)
        if "energy" in self.classical_calculator.results:
            potential_energy = self.classical_calculator.results["energy"]
        elif "potential_energy" in self.classical_calculator.results:
            potential_energy = self.classical_calculator.results["potential_energy"]
        else:
            potential_energy = potential_energies.sum()

        return forces, potential_energy, potential_energies

    def produce_qm_clusters(self, atoms,
                            potential_energies=None):
        """Update qm clusters based on potential energies per atom

        Parameters
        ----------
        atoms : ASE.atoms
            atoms object
        potential_energies : np.array
            Per atom potential energies
        """
        if self.forced_qm_list is None:
            # Use the newly calculated values to find the quantum mechanical regions
            self.cluster_list = self.qm_cluster.update_qm_region(atoms,
                                                                 potential_energies=potential_energies
                                                                 )
        else:
            if len(self.forced_qm_list) == 0:
                self.cluster_list = []
            else:
                self.cluster_list = self.forced_qm_list
        # Safeguard against empty clusters
        self.cluster_list = [item for item in self.cluster_list if len(item) > 0]

    def evaluate_qm_cluster_serial(self, atoms=None, cluster=None, clusterNumber=0):
        """Evaluate forces for a single QM cluster given the buffer hops

        Parameters
        ----------
        atoms : ASE.atoms
            atoms object
        cluster : list
            list of core qm atoms
        clusterNumber : int
            cluster number

        Returns
        -------
        Cluster : cluster_data
            object with forces
                        qm_atoms mark
                        core qm list
        """
        self.print_message("Evaluating cluster", limit=100)
        self.print_message(cluster, limit=100)

        # Create and evaluate hybrid cluster
        self.print_message("Creating cluster", limit=10)

        atomic_cluster = self.qm_cluster.carve_cluster(atoms, cluster, buffer_hops=self.buffer_hops)
        self.print_message("Size of the atomic cluster: " + str(len(atomic_cluster)), limit=10)

        # Debug cluster carving by printing the structures and clusters
        if (self.debug_cluster_carving):
            self.print_message("Writing cluster to file", limit=10)
            extension = "_" + str(clusterNumber + 1) + ".xyz"
            ase.io.write("cluster" + extension, atomic_cluster, format="xyz")
            ase.io.write("cluster_ext" + extension, atomic_cluster, format="extxyz")
            ase.io.write("structure.xyz", atoms, format="xyz")
            ase.io.write("structure_ext.xyz", atoms, format="extxyz", write_results=False)

            if (self.clusterDebug_cluster_atoms is not None):
                ase.io.write(self.clusterDebug_cluster_atoms, atomic_cluster,
                             format="extxyz", write_results=False)
            if (self.clusterDebug_full_structure is not None):
                ase.io.write(self.clusterDebug_full_structure, atoms,
                             format="extxyz", write_results=False)

        self.print_message("Evaluating", limit=100)
        qm_forces_array = self.qm_calculator.get_forces(atomic_cluster)

        self.print_message("qmmm pot cluster, " + str(len(atomic_cluster)) + " atoms long", limit=100)
        self.print_message(atomic_cluster.arrays["orig_index"], limit=100)

        # Create a cluster data object with relevan values
        mark = np.zeros(len(atoms), dtype=int)
        full_qm_forces = np.zeros((len(atoms), 3))

        for i in range(atomic_cluster.info["no_quantum_atoms"]):
            orig_index = atomic_cluster.arrays["orig_index"][i]
            full_qm_forces[orig_index, :] = qm_forces_array[i, :]
            mark[orig_index] = atomic_cluster.arrays["cluster_mark"][i]

        cluster_data = ClusterData(len(atoms), mark, cluster, full_qm_forces)

        # Try to add additional details to the cluster data
        qm_charges = np.zeros(len(atoms))
        if (self.debug_qm_calculator):
            try:
                # print('eigenvalues cluster:')
                # self.qm_calculator.print_eigenvalues(scope=15, offset=0)

                qm_charges -= 10
                atomic_cluster.arrays["qm_charges"] = self.qm_calculator.results["charges"].copy()
                atomic_cluster.arrays["qm_forces"] = self.qm_calculator.results["forces"].copy()

                ase.io.write("structure_ext.xyz", atoms, format="extxyz", write_results=False)
                ase.io.write("cluster_ext" + extension, atomic_cluster,
                             format="extxyz", write_results=False)

                for i in range(atomic_cluster.info["no_quantum_atoms"]):
                    orig_index = atomic_cluster.arrays["orig_index"][i]
                    qm_charges[orig_index] = self.qm_calculator.results["charges"][i]

            except KeyError:
                pass

        cluster_data.qm_charges = qm_charges
        cluster_data.nClusterAtoms = len(atomic_cluster)
        return cluster_data

    def combine_qm_mm_forces(self, atoms=None, forces=None, cluster_data_list=None):
        """This combines QM and MM forces

        Parameters
        ----------
        atoms : ASE.atoms
            atoms object
        forces : np.array
            atomic forces
        cluster_data_list : list of matscipy.calculators.mcfm.ClusterData
            information about the clusters

        Returns
        -------
        forces : np.array
            atomic forces
        """
        if (self.verbose >= 1):
            atoms.arrays["classical_forces"] = forces.copy()
        self.raw_qm_cluster_forces = np.zeros_like(forces)

        # If any clusters present, combine QM/MM forces, otherwise just pass forces along
        if ((self.long_range_weight > 0.0) and (len(cluster_data_list) > 0)):
            self.print_message(
                "Splitting forces into bonding/longRange and combining with QM if needed.", limit=10)

            # Obtain bonding forces
            forcesLR = self.classical_calculator.get_pairwise_forces(atoms)
            # Calculate long range forces
            # forcesLR = forces - forcesB

            # Replace bonding forces with QM forces for QM atoms only
            for cluster_data in cluster_data_list:
                for aI in cluster_data.qm_list:

                    # Combine short ranged forces with a fraction of the long range ones
                    # The long range forces are there for stability, should not affect the dynamics much.
                    forces[aI, :] = cluster_data.forces[aI, :] + forcesLR[aI, :] * self.long_range_weight
                    self.raw_qm_cluster_forces[aI, :] = cluster_data.forces[aI, :]

            # Add long range forces to the output
            if (self.verbose >= 1):
                atoms.arrays["Long_range_forces"] = forcesLR * self.long_range_weight

        elif (len(cluster_data_list) > 0):
            self.print_message("Combining QM and MM forces.", limit=10)
            for cluster_data in cluster_data_list:
                for aI in cluster_data.qm_list:
                    forces[aI, :] = cluster_data.forces[aI, :]
                    self.raw_qm_cluster_forces[aI, :] = cluster_data.forces[aI, :]

        if (self.verbose >= 1):
            atoms.arrays["raw_qm_forces"] = self.raw_qm_cluster_forces

        if (self.conserve_momentum) and ((len(cluster_data_list) > 0)):
            avg_force = forces.mean(axis=0)
            forces -= avg_force

        return forces

    def attach_hybrid_data(self, atoms=None, full_qm_atoms_mask=None, cluster_data=None):
        """Mark atoms that are treated quantm mmechanically
        for more comprehensive ooutput

        Parameters
        ----------
        atoms : ASE.atoms
            atoms object
        full_qm_atoms_mask : list
            list of all qm atoms
        cluster_data_list : list of matscipy.calculators.mcfm.ClusterData
            information about the clusters
        """

        # Store infrmation on individual clusters in atoms file
        atoms.arrays["hybrid_clusters"] = np.zeros(len(atoms))

        index = 0
        for cluster in self.cluster_list:
            # Safeguard against empty, nested lists
            if len(cluster) == 0:
                continue
            atoms.arrays["hybrid_clusters"][cluster] = index + 1
            atoms.arrays["cluster_marks_" + str(index + 1)] = cluster_data[index].mark.copy()

            # Add information about qm caharges
            if (self.verbose >= 1):
                atoms.arrays["qm_charges_clus_" + str(index + 1)] = cluster_data[index].qm_charges.copy()

            index += 1

    def evaluate_errors(self, atoms=None, heavy_only=False, r_force=None):
        """Use the forces and reference forces to get errors on hybrid atom
        force evaluations

        Parameters
        ----------
        atoms : ASE.atoms
            atoms object
        heavy_only : bool
            Do not evaluate errors on hydrogens
        r_force : np.array
            array with reference forces
        """

        # Create the full QM list and QM mask
        full_qm_atoms_list = [item for sublist in self.cluster_list for item in sublist]
        full_qm_atoms_mask = np.zeros(len(atoms), dtype=bool)
        full_qm_atoms_mask[full_qm_atoms_list] = True

        forces = self.raw_qm_cluster_forces
        if (r_force is None):
            if len(full_qm_atoms_list) > 0:
                r_force = self.qm_calculator.get_forces(atoms)
            else:
                r_force = forces

        atoms.arrays["reference_qm_force"] = r_force.copy()
        atoms.arrays["simulation_force"] = self.results["forces"].copy()
        atoms.arrays["qmmm_raw_force"] = self.raw_qm_cluster_forces.copy()

        try:
            if (len(self.qm_calculator.results["charges"]) == len(atoms) and (self.verbose >= 1)):
                atoms.arrays["reference_qm_charges"] = self.qm_calculator.results["charges"].copy()
        except KeyError:
            pass

        # Calculate errors for the QM regions
        # Only evaluate errors on heavy atoms if flag is set
        if (heavy_only is True):
            for i in range(len(full_qm_atoms_mask)):
                if (atoms.numbers[i] == 1):
                    full_qm_atoms_mask[i] = False

        if (full_qm_atoms_mask.sum() > 0):
            f_errorFull = r_force - forces
            f_error = np.linalg.norm(abs(f_errorFull), ord=2, axis=1)

            f_errFull = np.zeros((len(atoms), 3))
            f_errFull[full_qm_atoms_mask] = f_errorFull[full_qm_atoms_mask]
            f_err = np.zeros(len(atoms))
            f_err[full_qm_atoms_mask] = f_error[full_qm_atoms_mask]

            # Calculate if the errors are dumping or extrcting energy from the system
            # According to E = Force * velocity * timestep
            # Taking timestep as constant = 1
            try:
                energyChange = np.einsum("ij, ij -> i", f_errFull, atoms.arrays["momenta"])
                totalEChange = np.sum(energyChange)

                totalEnergyVector = np.einsum("ij, ij -> i", forces[full_qm_atoms_mask],
                                              atoms.arrays["momenta"][full_qm_atoms_mask])
                totalEnergy = np.sum(totalEnergyVector)

                totalEChange /= totalEnergy
            except KeyError:
                totalEChange = 0

            # Get the relative force error
            cumulative_forces = np.linalg.norm(forces, ord=2, axis=1)
            cumulative_forces = np.mean(cumulative_forces[full_qm_atoms_mask])
            relative_error = np.divide(f_err, cumulative_forces)

            max_relative_error = relative_error[full_qm_atoms_mask].max()
            rms_relative_error = np.sqrt(np.mean(np.square(relative_error[full_qm_atoms_mask])))

            max_absolute_error = f_err[full_qm_atoms_mask].max()
            rms_absolute_error = np.sqrt(np.mean(np.square(f_err[full_qm_atoms_mask])))

            # Provide max and RMS relative error
            # print "\tRMS of absolute errors: %0.5f, MAX absolute error: %0.5f" %
            # (rms_absolute_error, max_absolute_error)

        else:
            f_err = np.zeros(len(atoms))
            f_errFull = np.zeros((len(atoms), 3))
            relative_error = np.zeros(len(atoms))
            max_absolute_error = 0
            rms_absolute_error = 0
            max_relative_error = 0
            rms_relative_error = 0
            totalEChange = 0

        try:
            self.errors["Cumulative fError vector"] += f_errFull
            self.errors["Cumulative energy change"] += totalEChange
        except KeyError:
            self.errors["Cumulative fError vector"] = f_errFull
            self.errors["Cumulative energy change"] = totalEChange
        self.errors["vector force error"] = f_errFull
        self.errors["Cumulative fError vector length"] =\
            np.linalg.norm(abs(self.errors["Cumulative fError vector"]), ord=2, axis=1)
        self.errors["energy Change"] = totalEChange

        self.errors["absolute force error"] = f_err
        self.errors["relative force error"] = relative_error
        self.errors["max absolute error"] = max_absolute_error
        self.errors["rms absolute error"] = rms_absolute_error
        self.errors["max relative error"] = max_relative_error
        self.errors["rms relative error"] = rms_relative_error
        self.errors["rms force"] = np.sqrt(np.mean(np.square(forces)))
        self.errors["no of QM atoms"] = full_qm_atoms_mask.sum()

        # Add the relative error th the atoms object for visualization
        atoms.arrays["relative_Ferror"] = self.errors["relative force error"].copy()
        atoms.arrays["absolute_Ferror"] = self.errors["absolute force error"].copy()

        # #Calculate errors of the full system
        # f_err = np.linalg.norm(abs(r_force - forces), ord = 2, axis = 1)
        # relative_error = np.divide(f_err,
        #                     np.linalg.norm(forces, ord = 2, axis = 1))
        # max_relative_error = relative_error.max()
        # rms_relative_error = np.sqrt(np.mean(np.square(relative_error)))

        # max_absolute_error = f_err.max()
        # rms_absolute_error = np.sqrt(np.mean(np.square(f_err)))

        # self.errors_full = {}
        # self.errors_full["absolute force error"] = f_err
        # self.errors_full["relative force error"] = relative_error
        # self.errors_full["max absolute error"] = max_absolute_error
        # self.errors_full["rms absolute error"] = rms_absolute_error
        # self.errors_full["max relative error"] = max_relative_error
        # self.errors_full["rms relative error"] = rms_relative_error
        # self.errors_full["rms force"] = np.sqrt(np.mean(np.square(forces)))

    def set_qm_atoms(self, qm_list, atoms=None):
        """Force a certian set of clusters for qmmm evaluation,
        If forced_qm_list is assigned, the cluster list is not updated
        throughout the run

        Parameters
        ----------
        qm_list : list
            list of atoms
        atoms : ASE.atoms
            atoms object
        """
        if qm_list is None:
            self.forced_qm_list = None
        else:
            self.forced_qm_list = [qm_list]

    def compute_stress(self, atoms, forces):
        """Compute total stresses using viral theorem.
        WARNING: only works for non-PBC structures

        the formula for stress evaluation is
        ------------------------------------
        Sij = sum_k (m_k v_ik v_jk)/ volume + sum_k (r_ik f_jk)/volume
        m: mass
        v: velocity
        r: position
        f: force
        where i,j are taken from {x, y, z}
        and sum_k represents a sum over all atoms

        Parameters
        ----------
        atoms : ASE.atoms
            atoms object
        forces : np.array
            atomic forces

        Returns
        -------
        stress : np.array
            stress tensof in matrix notation
        """

        stress_mat = np.zeros((3, 3))
        stress = np.zeros(6)

        vol = atoms.get_volume()
        velo = atoms.get_velocities()
        mom = atoms.get_momenta()
        pos = atoms.get_positions()
        f = forces

        for i in range(3):
            for j in range(3):
                stress_mat[i, j] = - np.dot(pos[:, i], f[:, j]) / vol

        stress[0] = stress_mat[0, 0]
        stress[1] = stress_mat[1, 1]
        stress[2] = stress_mat[2, 2]
        stress[3] = stress_mat[1, 2]
        stress[4] = stress_mat[0, 2]
        stress[5] = stress_mat[0, 1]

        return stress

    def print_message(self, message, limit=100):
        """Print a message if the calculators verbosity level is above the
        given threshold

        For now verbose levels are
        --------------------------
        0 - nothing is printed
        1 - Message when meth::calculate is called
        10 - Calculate steps are listed
        100 - Information about specific QM clusters
        (the default is 0)

        Parameters
        ----------
        message : str
            The message to be printed
        limit : int
            the verbosity threshold for this mesage
        """
        if (self.verbose >= limit):
            print(message)



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014-2017) James Kermode, Warwick University
#                       Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

"""
Tools related to homogenously deformed volumes.
"""

import numpy as np

###

def get_shear_distance(a):
    """
    Returns the distance a volume has moved during simple shear. Considers
    either Lees-Edwards boundary conditions or sheared cells.
    """
    cx, cy, cz = a.cell
    if 'shear_dx' in a.info:
        assert abs(cx[1]) < 1e-12, 'cx[1] = {0}'.format(cx[1])
        assert abs(cx[2]) < 1e-12, 'cx[2] = {0}'.format(cx[2])
        assert abs(cy[0]) < 1e-12, 'cx[0] = {0}'.format(cy[0])
        assert abs(cy[2]) < 1e-12, 'cy[2] = {0}'.format(cy[2])
        assert abs(cz[0]) < 1e-12, 'cz[0] = {0}'.format(cz[0])
        assert abs(cz[1]) < 1e-12, 'cz[1] = {0}'.format(cz[1])
        dx, dy, dz = a.info['shear_dx']
    else:
        assert abs(cx[1]) < 1e-12, 'cx[1] = {0}'.format(cx[1])
        assert abs(cx[2]) < 1e-12, 'cx[2] = {0}'.format(cx[2])
        assert abs(cy[0]) < 1e-12, 'cy[0] = {0}'.format(cy[0])
        assert abs(cy[2]) < 1e-12, 'cy[2] = {0}'.format(cy[2])
        dx, dy, sz = cz
    return dx, dy

###

class RemoveSimpleShearDeformation:
    """
    Remove a homogeneous cell deformation given an (iterable) trajectory
    object. This will take proper care of cells that are instantaneously
    flipped from +0.5 strain to -0.5 strain during simple shear, as e.g.
    generated by LAMMPS.
    """

    def __init__(self, traj):
        self.traj = traj

        self.last_d = [ ]

        self.sheared_cells = [ ]
        self.unsheared_cells = [ ]


    def _fill_cell_info_upto(self, i):
        # Iterate up to frame i the full trajectory first and generate a list
        # of cell vectors.
        if i < len(self.last_d):
            return

        # Iterate up to frame i the full trajectory first and generate a list
        # of cell vectors.
        if len(self.last_d) == 0:
            i0 = 0
            last_dx, last_dy = get_shear_distance(self.traj[0])
            dx = last_dx
            dy = last_dy
        else:
            i0 = len(self.last_d)
            last_dx, last_dy = self.last_d[i0-1]
            dx, dy, dummy = self.sheared_cells[i0-1][2]

        for a in self.traj[i0:i+1]:
            sx, sy, sz = a.cell.diagonal()
            cur_dx, cur_dy = get_shear_distance(a)
            while cur_dx-last_dx < -sx/2:
                cur_dx += sx
            dx += cur_dx-last_dx
            while cur_dy-last_dy < -sy/2:
                cur_dy += sy
            dy += cur_dy-last_dy

            # Store last shear distance
            last_dx = cur_dx
            last_dy = cur_dy

            # Store cells and shear distance
            self.last_d += [ ( last_dx, last_dy ) ]

            self.sheared_cells += [ np.array([[sx,0,0],[0,sy,0],[dx,dy,sz]]) ]
            self.unsheared_cells += [ np.array([sx,sy,sz]) ]


    def __getitem__(self, i=-1):
        if i < 0:
            i = len(self) + i
            if i < 0 or i >= len(self):
                raise IndexError('Trajectory index out of range.')

        self._fill_cell_info_upto(i)

        a = self.traj[i]

        # Set true cell shape
        a.set_cell(self.sheared_cells[i], scale_atoms=False)
        # Unshear
        a.set_cell(self.unsheared_cells[i], scale_atoms=True)

        # Wrap to cell
        a.set_scaled_positions(a.get_scaled_positions()%1.0)

        a.info['true_cell'] = self.sheared_cells[i]

        return a


    def __len__(self):
        return len(self.traj)

